# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for QiskitProvider class

"""
import random
import warnings
from collections import OrderedDict
from types import SimpleNamespace
from unittest.mock import MagicMock, Mock, patch

import numpy as np
import pytest
from qiskit import QuantumCircuit
from qiskit.exceptions import QiskitError
from qiskit.primitives.primitive_job import PrimitiveJob
from qiskit.providers.fake_provider import GenericBackendV2
from qiskit.transpiler.preset_passmanagers import generate_preset_pass_manager
from qiskit_ibm_runtime import QiskitRuntimeService, RuntimeJob
from qiskit_ibm_runtime.exceptions import IBMNotAuthorizedError, RuntimeInvalidStateError
from qiskit_ibm_runtime.qiskit_runtime_service import QiskitBackendNotFoundError

from qbraid.programs import NATIVE_REGISTRY, ExperimentType, ProgramSpec
from qbraid.runtime import DeviceStatus, GateModelResultData, JobStateError, Result, TargetProfile
from qbraid.runtime.exceptions import QbraidRuntimeError
from qbraid.runtime.ibm import QiskitBackend, QiskitJob, QiskitRuntimeProvider
from qbraid.runtime.ibm.result_builder import QiskitGateModelResultBuilder

try:
    from qiskit.providers.models.backendconfiguration import QasmBackendConfiguration

except ModuleNotFoundError:
    from qiskit_ibm_runtime.models.backend_configuration import QasmBackendConfiguration

FIXTURE_COUNT = sum(key in NATIVE_REGISTRY for key in ["qiskit", "braket", "cirq"])

FAKE_BASIS_GATES = ["u1", "u2", "u3", "cx"]


class FakeDevice(GenericBackendV2):
    """A test Qiskit device."""

    # pylint: disable-next=too-many-arguments
    def __init__(
        self,
        num_qubits,
        local=True,
        simulator=True,
        operational=True,
        status_msg="active",
        **kwargs,
    ):
        super().__init__(num_qubits)
        self._num_qubits = num_qubits
        self._operational = operational
        self._status_msg = status_msg
        self._local = local
        self._simulator = simulator
        self._instance = kwargs.get("instance", None)

    def status(self):
        """Return the status of the backend."""
        status_obj = SimpleNamespace()
        status_obj.operational = self._operational
        status_obj.status_msg = self._status_msg

        # dummy value proportional to num_qubits
        status_obj.pending_jobs = 0 if self._local else random.randint(1, 100)
        return status_obj

    def configuration(self):
        """Return the configuration of the backend."""
        return QasmBackendConfiguration(
            backend_name="fake_backend",
            backend_version="1.0",
            n_qubits=self._num_qubits,
            basis_gates=FAKE_BASIS_GATES,
            gates=[],
            local=self._local,
            simulator=self._simulator,
            conditional=False,
            open_pulse=False,
            memory=True,
            max_shots=8192,
            coupling_map=None,
        )


class FakeService:
    """Fake Qiskit runtime service for testing."""

    def backend(self, backend_name, instance=None):
        """Return fake backend."""
        for backend in self.backends(instance=instance):
            if backend_name == backend.name:
                return backend
        raise QiskitBackendNotFoundError("No backend matches the criteria.")

    def backends(self, **kwargs):  # pylint: disable=unused-argument
        """Return fake Qiskit backend."""
        return [FakeDevice(num_qubits=n, **kwargs) for n in [5, 20]]

    def backend_names(self, **kwargs):
        """Return fake backend names."""
        return [backend.name for backend in self.backends(**kwargs)]

    def least_busy(self, **kwargs):
        """Return the least busy backend based on the number of pending jobs."""
        backends = self.backends(**kwargs)
        return min(backends, key=lambda backend: backend.status().pending_jobs)

    def job(self, job_id):  # pylint: disable=unused-argument
        """Return fake job."""
        return MagicMock(spec=RuntimeJob)


def _create_backend_fixture(service: FakeService, local: bool, simulator: bool) -> QiskitBackend:
    """Create a Qiskit backend fixture."""
    backend = service.backends(local=local, simulator=simulator)[0]
    program_spec = ProgramSpec(QuantumCircuit)
    profile = TargetProfile(
        device_id=backend.name,
        local=local,
        simulator=simulator,
        experiment_type=ExperimentType.GATE_MODEL,
        num_qubits=backend.num_qubits,
        program_spec=program_spec,
        provider_name="IBM",
    )
    return QiskitBackend(profile, service)


@pytest.fixture
def fake_service():
    """Return a fake Qiskit runtime service."""
    return FakeService()


@pytest.fixture
def fake_device(fake_service):
    """Return a fake local simulator backend."""
    return _create_backend_fixture(fake_service, local=True, simulator=True)


def test_provider_initialize_service():
    """Test getting IBMQ provider using qiskit_ibm_provider package."""
    with patch("qbraid.runtime.ibm.provider.QiskitRuntimeService") as mock_runtime_service:
        mock_runtime_service.return_value = Mock(spec=QiskitRuntimeService)
        provider = QiskitRuntimeProvider(token="test_token", channel="test_channel")
        assert isinstance(provider.runtime_service, QiskitRuntimeService)
        assert provider.token == "test_token"
        assert provider.channel == "test_channel"
        assert provider.runtime_service == mock_runtime_service.return_value


def test_provider_save_config(fake_service):
    """Test saving the IBM account configuration to disk."""
    with patch("qbraid.runtime.ibm.provider.QiskitRuntimeService") as mock_runtime_service:
        mock_runtime_service.return_value = fake_service
        provider = QiskitRuntimeProvider(
            token="fake_token", instance="ibm-q/open/main", channel="fake_channel"
        )

        provider.save_config(
            token="fake_token", instance="ibm-q/open/main", channel="fake_channel", overwrite=False
        )
        mock_runtime_service.save_account.assert_called_once_with(
            token="fake_token", instance="ibm-q/open/main", channel="fake_channel", overwrite=False
        )

        provider.save_config()
        mock_runtime_service.save_account.assert_called_with(
            token="fake_token", instance="ibm-q/open/main", channel="fake_channel", overwrite=True
        )


@pytest.mark.parametrize(
    "local,simulator,num_qubits",
    [
        (True, True, 20),
        (False, True, 5),
        (False, False, 5),
    ],
)
def test_provider_build_runtime_profile(local, simulator, num_qubits):
    """Test building runtime profile for Qiskit backend."""
    with patch("qbraid.runtime.ibm.provider.QiskitRuntimeService") as mock_runtime_service:
        mock_runtime_service.return_value = FakeService()
        backend = FakeDevice(num_qubits, local=local, simulator=simulator)
        provider = QiskitRuntimeProvider(token="test_token", channel="test_channel")
        profile = provider._build_runtime_profile(backend)
        assert profile["device_id"] == f"generic_backend_{num_qubits}q"
        assert profile["simulator"] == simulator
        assert profile["local"] == local
        assert profile["num_qubits"] == num_qubits
        assert profile["max_shots"] == 8192
        assert profile.basis_gates == set(FAKE_BASIS_GATES)

        qiskit_backend = QiskitBackend(profile, mock_runtime_service())
        assert isinstance(qiskit_backend, QiskitBackend)
        assert qiskit_backend.profile == profile


def test_provider_get_devices(fake_service):
    """Test getting a backend from the runtime service."""
    with patch("qbraid.runtime.ibm.provider.QiskitRuntimeService") as mock_runtime_service:
        mock_runtime_service.return_value = fake_service
        provider = QiskitRuntimeProvider(token="test_token", channel="test_channel")
        devices = provider.get_devices()
        assert len(devices) == 2
        assert all(isinstance(device, QiskitBackend) for device in devices)

        device = devices[0]
        device_copy = provider.get_device(device.id)
        assert device.id == device_copy.id
        assert str(device) == f"QiskitBackend('{device.id}')"


@pytest.mark.parametrize("local", [True, False])
def test_provider_least_busy(fake_service, local):
    """Test getting a backend from the runtime service, both local and non-local."""
    with patch("qbraid.runtime.ibm.provider.QiskitRuntimeService") as mock_provider_service:
        mock_provider_service.return_value = fake_service
        provider = QiskitRuntimeProvider(token="test_token", channel="test_channel")

        least_busy = provider.least_busy(local=local)
        least_busy._backend = FakeDevice(5, local=local)

        assert least_busy.queue_depth() == 0 if local else least_busy.queue_depth() > 0


@pytest.mark.parametrize(
    "operational,local,status_msg,expected_status",
    [
        (True, True, "active", DeviceStatus.ONLINE),
        (True, False, "active", DeviceStatus.ONLINE),
        (True, False, "not active", DeviceStatus.UNAVAILABLE),
        (False, False, None, DeviceStatus.OFFLINE),
    ],
)
def test_device_status(fake_service, operational, local, status_msg, expected_status):
    """Test getting a backend from the runtime service, both local and non-local."""
    with patch("qbraid.runtime.ibm.provider.QiskitRuntimeService") as mock_provider_service:
        mock_provider_service.return_value = fake_service
        provider = QiskitRuntimeProvider(token="test_token", channel="test_channel")
        params = {"operational": operational, "local": local, "status_msg": status_msg}
        device = provider.get_devices(**params)[0]
        device._backend = FakeDevice(device.num_qubits, **params)
        assert device.status() == expected_status


@pytest.mark.parametrize("circuit_meas", range(FIXTURE_COUNT), indirect=True)
def test_device_run(fake_device, circuit_meas):
    """Test run method from wrapped fake Qiskit backends"""
    with warnings.catch_warnings():
        warnings.simplefilter(action="ignore", category=RuntimeWarning)
        qbraid_job = fake_device.run(circuit_meas, shots=10)

    vendor_job = qbraid_job._job
    assert isinstance(qbraid_job, QiskitJob)
    assert isinstance(vendor_job, PrimitiveJob)


def test_device_run_batch(fake_device, run_inputs):
    """Test run method from wrapped fake Qiskit backends"""
    with warnings.catch_warnings():
        warnings.simplefilter(action="ignore", category=RuntimeWarning)
        qbraid_job = fake_device.run(run_inputs, shots=10)

    vendor_job = qbraid_job._job
    assert isinstance(qbraid_job, QiskitJob)
    assert isinstance(vendor_job, PrimitiveJob)


@pytest.fixture
def mock_service():
    """Fixture to create a mock QiskitRuntimeService."""
    service = MagicMock(spec=QiskitRuntimeService)
    service.job.return_value = MagicMock(spec=RuntimeJob)
    return service


def test_job_initialize_from_service(mock_service):
    """Test successful job retrieval with a provided service.

    Args:
        mock_service (MagicMock): A mocked service passed as a fixture.

    Tests that QiskitJob retrieves a job using a provided mock service and verifies
    that the job method is called exactly once with the correct job ID.
    """
    job_id = "test_job_id"
    job = QiskitJob(job_id, service=mock_service)
    assert job._job is not None
    mock_service.job.assert_called_once_with(job_id)


def test_job_initialize_service_from_device(mock_service):
    """Test job retrieval when service is provided via the device attribute."""
    mock_device = MagicMock()
    mock_device._service = mock_service

    job_id = "test_job_id"
    job = QiskitJob(job_id, device=mock_device)

    assert job._job is not None
    mock_service.job.assert_called_once_with(job_id)


def test_job_service_initialization():
    """Test job retrieval when initializing a new service."""
    with patch("qbraid.runtime.ibm.provider.QiskitRuntimeService") as mock_service_class:
        mock_service = MagicMock(spec=QiskitRuntimeService)
        mock_service.job.return_value = MagicMock(spec=RuntimeJob)
        mock_service_class.return_value = mock_service
        job_id = "test_job_id"
        job = QiskitJob(job_id)
        assert job._job is not None
        mock_service.job.assert_called_once_with(job_id)


def test_job_service_initialization_failure():
    """Test handling of service initialization failure."""
    with patch(
        "qbraid.runtime.ibm.provider.QiskitRuntimeService", side_effect=IBMNotAuthorizedError
    ):
        job_id = "test_job_id"
        with pytest.raises(QbraidRuntimeError) as excinfo:
            QiskitJob(job_id)
        assert "Failed to initialize the quantum service." in str(excinfo.value)


def test_job_retrieval_failure(mock_service):
    """Test handling of job retrieval failure."""
    mock_service.job.side_effect = ConnectionError
    job_id = "test_job_id"
    with pytest.raises(QbraidRuntimeError) as excinfo:
        QiskitJob(job_id, service=mock_service)
    assert "Error retrieving job test_job_id" in str(excinfo.value)


@pytest.fixture
def mock_runtime_job():
    """Fixture to create a mock QiskitRuntimeJob object."""
    return MagicMock()


def test_job_queue_position(mock_runtime_job):
    """Test that the queue position of the job is correctly returned."""
    queue_position = 5
    mock_runtime_job.queue_position.return_value = queue_position

    job = QiskitJob(job_id="123", job=mock_runtime_job)
    assert job.queue_position() == queue_position


def test_cancel_job_in_terminal_state(mock_runtime_job):
    """Test attempting to cancel a job in a terminal state raises JobStateError."""
    job = QiskitJob(job_id="123", job=mock_runtime_job)
    job.is_terminal_state = MagicMock(return_value=True)  # Simulate terminal state

    with pytest.raises(JobStateError, match="Cannot cancel quantum job in non-terminal state."):
        job.cancel()


def test_cancel_job_success(mock_runtime_job):
    """Test successful cancellation of a non-terminal job."""
    job = QiskitJob(job_id="123", job=mock_runtime_job)
    job.is_terminal_state = MagicMock(return_value=False)
    mock_runtime_job.cancel.return_value = None

    assert job.cancel() is None


def test_cancel_job_fails_due_to_runtime_error(mock_runtime_job):
    """Test that cancellation fails with a RuntimeInvalidStateError and raises JobStateError."""
    mock_job = mock_runtime_job
    mock_job.cancel.side_effect = RuntimeInvalidStateError
    job = QiskitJob(job_id="123", job=mock_job)
    job.is_terminal_state = MagicMock(return_value=False)

    with pytest.raises(JobStateError):
        job.cancel()


@pytest.fixture
def mock_result_metadata():
    """Return a dictionary of mock result metadata."""
    return {
        "job_id": "123",
        "success": True,
        "backend_name": "fake_backend",
        "backend_version": "0.0.0",
    }


@pytest.fixture
def mock_runtime_result(mock_result_metadata):
    """Return a mock Qiskit result."""
    mock_result = Mock()
    mock_result.results = [Mock(), Mock()]
    meas1 = ["01"] * 9 + ["10"] + ["11"] * 4 + ["00"] * 6
    meas2 = ["0"] * 8 + ["1"] * 12
    mock_result.get_memory.side_effect = [meas1, meas2]
    mock_result.get_counts.side_effect = [{"01": 9, "10": 1, "11": 4, "00": 6}, {"0": 8, "1": 12}]
    mock_result.to_dict.return_value = mock_result_metadata
    return mock_result


def test_result_from_job(mock_runtime_job, mock_runtime_result, mock_result_metadata):
    """Test that result returns QiskitGateModelResultBuilder when the job is in a terminal state."""
    mock_job = QiskitJob(job_id="123", job=mock_runtime_job)
    mock_runtime_job.result.return_value = mock_runtime_result
    result = mock_job.result()
    assert isinstance(result, Result)
    assert isinstance(result.data, GateModelResultData)
    assert isinstance(result.data.get_counts(), dict)
    assert isinstance(result.data.measurements, np.ndarray)
    assert result.details.get("job_id") is None
    assert result.details.get("success") is None
    assert result.details.get("backend_name") == mock_result_metadata["backend_name"]
    assert result.details.get("backend_version") == mock_result_metadata["backend_version"]


def test_result_get_counts(mock_runtime_result):
    """Test getting raw counts from a Qiskit result."""
    qr = QiskitGateModelResultBuilder(mock_runtime_result)
    expected = {"01": 9, "10": 1, "11": 4, "00": 6}
    assert qr.get_counts() == expected


def test_result_format_measurements(mock_runtime_result):
    """Test formatting measurements into integers."""
    qr = QiskitGateModelResultBuilder(mock_runtime_result)
    memory_list = ["010", "111"]
    expected = [[0, 1, 0], [1, 1, 1]]
    assert qr._format_measurements(memory_list) == expected


def test_result_measurements_single_circuit(mock_runtime_result):
    """Test getting measurements from a single circuit."""
    qr = QiskitGateModelResultBuilder(mock_runtime_result)
    mock_runtime_result.results = [Mock()]
    expected = np.array([[0, 1]] * 9 + [[1, 0]] + [[1, 1]] * 4 + [[0, 0]] * 6)
    assert np.array_equal(qr.measurements(), expected)


def test_result_measurements_multiple_circuits(mock_runtime_result):
    """Test getting measurements from multiple circuits."""
    qr = QiskitGateModelResultBuilder(mock_runtime_result)
    expected_meas1 = np.array([[0, 1]] * 9 + [[1, 0]] + [[1, 1]] * 4 + [[0, 0]] * 6)
    expected_meas2 = np.array([[0, 0]] * 8 + [[0, 1]] * 12)
    expected = np.array([expected_meas1, expected_meas2])
    assert np.array_equal(qr.measurements(), expected)


@patch("builtins.hash", autospec=True)
def test_hash_method_creates_and_returns_hash(mock_hash):
    """Test that the __hash__ method creates and returns a hash value."""
    with patch("qbraid.runtime.ibm.provider.QiskitRuntimeService") as mock_runtime_service:
        mock_runtime_service.return_value = Mock(spec=QiskitRuntimeService)
        provider_instance = QiskitRuntimeProvider(token="mock_token", channel="ibm_cloud")
        mock_hash.return_value = 12345
        result = provider_instance.__hash__()  # pylint: disable=unnecessary-dunder-call
        mock_hash.assert_called_once_with(("mock_token", "ibm_cloud"))
        assert result == 12345
        assert provider_instance._hash == 12345


def test_hash_method_returns_existing_hash():
    """Test that the hash method returns the existing hash value."""
    with patch("qbraid.runtime.ibm.provider.QiskitRuntimeService") as mock_runtime_service:
        mock_runtime_service.return_value = Mock(spec=QiskitRuntimeService)
        provider_instance = QiskitRuntimeProvider(token="mock_token", channel="ibm_cloud")
        provider_instance._hash = 67890
        result = provider_instance.__hash__()  # pylint: disable=unnecessary-dunder-call
        assert result == 67890


def test_result_measurements_not_available(mock_runtime_result, caplog):
    """Test returning None and logging a warning when measurements are not available."""
    qr = QiskitGateModelResultBuilder(mock_runtime_result)
    mock_result = Mock()
    mock_result.results = qr._result.results
    mock_result.get_memory.side_effect = QiskitError("Test QiskitError message")
    qr._result = mock_result

    with caplog.at_level("WARNING"):
        result = qr.measurements()

    assert result is None
    assert "Memory states (measurements) data not available for this job" in caplog.text
    assert "Test QiskitError message" in caplog.text
    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == "WARNING"
    assert (
        "Memory states (measurements) data not available for this job" in caplog.records[0].message
    )


def test_transform_run_input(fake_device):
    """Test that the transform method optimizes the circuit for the device."""
    circuit = QuantumCircuit(4)

    circuit.h(0)
    circuit.cx(0, 1)
    circuit.cx(1, 2)
    circuit.cx(2, 3)
    circuit.measure_all()

    circuit_optimized_for_device = fake_device.transform(circuit)

    pm = generate_preset_pass_manager()

    fake_device.set_options(pass_manager=pm)

    circuit_optimized_default = fake_device.transform(circuit)

    default_ops = OrderedDict([("measure", 4), ("cx", 3), ("h", 1), ("barrier", 1)])
    optimized_ops = OrderedDict([("measure", 4), ("cx", 3), ("rz", 2), ("sx", 1), ("barrier", 1)])

    assert circuit.count_ops() == circuit_optimized_default.count_ops() == default_ops
    assert circuit_optimized_for_device.count_ops() == optimized_ops


def test_device_run_with_sampler(fake_device, qiskit_uniform):
    """Test that the device can run a circuit with a sampler."""
    shots = 10
    num_qubits = 2

    qbraid_job = fake_device.run(qiskit_uniform(num_qubits), shots=shots)
    assert isinstance(qbraid_job, QiskitJob)

    result = qbraid_job.result()
    assert isinstance(result, Result)
    assert isinstance(result.data, GateModelResultData)
    assert isinstance(result.data.get_counts(), dict)
    assert isinstance(result.data.measurements, np.ndarray)
    assert result.data.measurements.shape == (shots, 1)

    counts = result.data.get_counts()
    assert sum(counts.values()) == shots
    assert all(len(k) == num_qubits for k in counts.keys())



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for generating random OpenQASM 3 programs

"""
from typing import Callable, Optional

import numpy as np

from qbraid._version import __version__
from qbraid.exceptions import QbraidError


def create_gateset_qasm(max_operands: int) -> np.ndarray:
    """Gets QASM for gateset with max_operands."""
    q1_gates: list[tuple[str, int, int]] = [
        ("id", 1, 0),
        ("x", 1, 0),
        ("y", 1, 0),
        ("z", 1, 0),
        ("h", 1, 0),
        ("s", 1, 0),
        ("t", 1, 0),
        ("sdg", 1, 0),
        ("tdg", 1, 0),
        ("sx", 1, 0),
        ("rx", 1, 1),
        ("ry", 1, 1),
        ("rz", 1, 1),
        ("p", 1, 1),
        ("u1", 1, 1),
        ("u2", 1, 2),
        ("u3", 1, 3),
        ("reset", 1, 0),
    ]

    q2_gates: list[tuple[str, int, int]] = [
        ("cx", 2, 0),
        ("cy", 2, 0),
        ("cz", 2, 0),
        ("ch", 2, 0),
        ("cp", 2, 1),
        ("crx", 2, 1),
        ("cry", 2, 1),
        ("crz", 2, 1),
        ("swap", 2, 0),
        ("cu", 2, 4),
    ]

    q3_gates: list[tuple[str, int, int]] = [("ccx", 3, 0), ("cswap", 3, 0)]

    gates = q1_gates.copy()

    if max_operands >= 2:
        gates.extend(q2_gates)
    if max_operands >= 3:
        gates.extend(q3_gates)

    gates_array = np.array(
        gates, dtype=[("gate", object), ("num_qubits", np.int64), ("num_params", np.int64)]
    )
    return gates_array


# pylint: disable-next=too-many-arguments
def qasm3_random_from_gates(
    create_gateset: Callable[[int], np.ndarray],
    num_qubits: Optional[int] = None,
    depth: Optional[int] = None,
    max_operands: Optional[int] = None,
    seed: Optional[int] = None,
    measure: bool = False,
) -> str:
    """Generate random OpenQASM 3 program.

    Args:
        create_gateset (Callable): Function to create gateset.
        num_qubits (int): Number of quantum wires.
        depth (int): Layers of operations (i.e., critical path length).
        max_operands (int): Maximum size of gate for each operation.
        seed (int): Seed for random number generator.
        measure (bool): Whether to include measurement gates.

    Raises:
        ValueError: When invalid random circuit options are given.
        QbraidError: When failed to create random OpenQASM 3 program.

    Returns:
        str: OpenQASM 3 program.
    """

    def validate_and_assign(value: Optional[int], name: str) -> int:
        """Validate and assign random circuit option."""
        if value is None:
            return np.random.randint(1, 4)
        if not isinstance(value, int) or value <= 0:
            raise ValueError(f"Invalid random circuit option. '{name}' must be a positive integer.")
        return value

    num_qubits = validate_and_assign(num_qubits, "num_qubits")
    depth = validate_and_assign(depth, "depth")
    max_operands = validate_and_assign(max_operands, "max_operands")

    try:
        if seed is None:
            seed = np.random.randint(0, np.iinfo(np.int32).max)
        np.random.seed(seed)
        rng = np.random.default_rng(seed)

        qasm_code_header = f"""
// Generated by qBraid v{__version__}
OPENQASM 3.0;
include "stdgates.inc";
/*
    seed = {seed}
    num_qubits = {num_qubits}
    depth = {depth}
    max_operands = {max_operands}
*/
"""
        max_operands = min(max_operands, num_qubits)
        rand_circuit = qasm_code_header + f"qubit[{num_qubits}] q;\n"
        if measure:
            rand_circuit += f"bit[{num_qubits}] c;\n"

        qubits = np.arange(num_qubits)
        gates = create_gateset(max_operands)

        for _ in range(depth):
            gate_specs = rng.choice(gates, size=num_qubits)
            cumulative_qubits = np.cumsum(gate_specs["num_qubits"], dtype=np.int64)

            max_index = np.searchsorted(cumulative_qubits, num_qubits, side="right")
            gate_specs = gate_specs[:max_index]
            slack = num_qubits - cumulative_qubits[max_index - 1]
            if slack:
                gates = create_gateset(1)
                slack_gates = rng.choice(gates, size=slack)
                gate_specs = np.hstack((gate_specs, slack_gates))

            q_indices = np.empty(len(gate_specs) + 1, dtype=np.int64)
            p_indices = np.empty(len(gate_specs) + 1, dtype=np.int64)
            q_indices[0] = p_indices[0] = 0
            np.cumsum(gate_specs["num_qubits"], out=q_indices[1:])
            np.cumsum(gate_specs["num_params"], out=p_indices[1:])
            parameters = rng.uniform(0, 2 * np.pi, size=p_indices[-1])
            for i, (gate, _, p) in enumerate(gate_specs):
                if p:
                    params = ",".join(
                        str(parameters[j]) for j in range(p_indices[i], p_indices[i + 1])
                    )
                    qubit_indices = ",".join(
                        f"q[{qubits[j]}]" for j in range(q_indices[i], q_indices[i + 1])
                    )
                    rand_circuit += f"{gate}({params}) {qubit_indices};\n"
                else:
                    qubit_indices = ",".join(
                        f"q[{qubits[j]}]" for j in range(q_indices[i], q_indices[i + 1])
                    )
                    rand_circuit += f"{gate} {qubit_indices};\n"
            qubits = rng.permutation(qubits)

        if measure:
            for i in range(num_qubits):
                rand_circuit += f"c[{i}] = measure q[{i}];\n"

        return rand_circuit

    except Exception as err:
        raise QbraidError("Failed to create random OpenQASM 3 program") from err


def qasm3_random(
    num_qubits: Optional[int] = None,
    depth: Optional[int] = None,
    max_operands: Optional[int] = None,
    seed: Optional[int] = None,
    measure: bool = False,
) -> str:
    """Generate random OpenQASM 3 program.

    Args:
        num_qubits (int): Number of quantum wires.
        depth (int): Layers of operations (i.e., critical path length).
        max_operands (int): Maximum size of gate for each operation.
        seed (int): Seed for random number generator.
        measure (bool): Whether to include measurement gates.

    Raises:
        ValueError: When invalid random circuit options are given.
        QbraidError: When failed to create random OpenQASM 3 program.

    Returns:
        str: OpenQASM 3 program.
    """
    return qasm3_random_from_gates(
        create_gateset_qasm, num_qubits, depth, max_operands, seed, measure
    )



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for Cirq gate dictionary used for testing

"""
import numpy as np
from cirq.ops.common_gates import (
    CXPowGate,
    CZPowGate,
    HPowGate,
    XPowGate,
    YPowGate,
    ZPowGate,
    cphase,
    rx,
    ry,
    rz,
)
from cirq.ops.identity import IdentityGate
from cirq.ops.matrix_gates import MatrixGate
from cirq.ops.swap_gates import ISwapPowGate, SwapPowGate
from cirq.ops.three_qubit_gates import CCXPowGate, CCZPowGate

from qbraid.transpiler.conversions.braket.braket_to_cirq import _give_cirq_gate_name
from qbraid.transpiler.conversions.qasm2.cirq_custom import U3Gate

cirq_gates = {
    "H": HPowGate,
    "X": XPowGate,
    "Y": YPowGate,
    "Z": ZPowGate,
    "S": ZPowGate,
    "T": ZPowGate,
    "I": IdentityGate,
    "HPow": HPowGate,
    "XPow": XPowGate,
    "YPow": YPowGate,
    "ZPow": ZPowGate,
    "RX": rx,
    "RY": ry,
    "RZ": rz,
    "Phase": ZPowGate,
    "U1": ZPowGate,
    "CX": CXPowGate,
    "Swap": SwapPowGate,
    "iSwap": ISwapPowGate,
    "CZ": CZPowGate,
    "CPhase": cphase,
    "CCZ": CCZPowGate,
    "CCX": CCXPowGate,
    "U3": U3Gate,
}


def create_cirq_gate(data):
    gate_type = data["type"]
    params = data["params"]
    matrix = data["matrix"]

    # single-qubit, no parameters
    if gate_type in ("H", "X", "Y", "Z"):
        return cirq_gates[gate_type]()

    elif gate_type == "I":
        return cirq_gates["I"](num_qubits=1)

    elif gate_type == "S":
        return cirq_gates["S"](exponent=0.5)
    elif gate_type == "T":
        return cirq_gates["T"](exponent=0.25)

    # single-qubit, one-parameter gates
    elif gate_type in ("RX", "RY", "RZ"):
        theta = data["params"][0]
        # theta = data["params"][0] / np.pi
        return cirq_gates[gate_type](theta)

    elif gate_type in ("HPow", "XPow", "YPow", "ZPow"):
        exponent = data["params"][0]
        return cirq_gates[gate_type](exponent=exponent)

    elif gate_type in ("Phase", "U1"):
        t = data["params"][0] / np.pi
        return cirq_gates["Phase"](exponent=t)

    # two-qubit, no parameters
    elif gate_type in ("CX", "CZ"):
        return cirq_gates[gate_type]()  # default exponent = 1

    elif gate_type in "CPhase":
        return cirq_gates[gate_type](data["params"][0])

    elif gate_type in ("Swap", "iSwap"):
        return cirq_gates[gate_type](exponent=1.0)

    # multi-qubit
    elif gate_type in "CCX":
        return cirq_gates[gate_type]()

    # custom gates
    elif gate_type == "U3":
        return U3Gate(*params)

    elif gate_type == "Unitary" or matrix is not None:
        n_qubits = int(np.log2(len(matrix)))
        unitary_gate = MatrixGate(matrix)
        _give_cirq_gate_name(unitary_gate, "U", n_qubits)
        return unitary_gate

    else:
        raise ValueError(f"Gate of type {gate_type} not supported for Cirq testing.")



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Device class for OQC devices.

"""
from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Optional, Union

import pyqasm
from qcaas_client.client import QPUTask
from qcaas_client.compiler_config import (
    CompilerConfig,
    MetricsType,
    QuantumResultsFormat,
    Tket,
    TketOptimizations,
)

from qbraid._logging import logger
from qbraid.runtime.device import QuantumDevice
from qbraid.runtime.enums import DeviceStatus
from qbraid.runtime.exceptions import ResourceNotFoundError

from .job import OQCJob

if TYPE_CHECKING:
    import qcaas_client.client

    import qbraid.runtime

RESULTS_FORMAT = {
    "binary": QuantumResultsFormat().binary_count(),
    "raw": QuantumResultsFormat().raw(),
}

METRICS_TYPE = {
    "default": MetricsType.Default,
    "empty": MetricsType.Empty,
    "optimized_circuit": MetricsType.OptimizedCircuit,
    "optimized_instruction_count": MetricsType.OptimizedInstructionCount,
}

OPTIMIZATIONS = {
    "clifford_simplify": Tket(TketOptimizations.CliffordSimp),
    "context_simplify": Tket(TketOptimizations.ContextSimp),
    "decompose_controlled_gates": Tket(TketOptimizations.DecomposeArbitrarilyControlledGates),
    "default_mapping_pass": Tket(TketOptimizations.DefaultMappingPass),
    "empty": Tket(TketOptimizations.Empty),
    "full_peephole_optimise": Tket(TketOptimizations.FullPeepholeOptimise),
    "globalise_phased_x": Tket(TketOptimizations.GlobalisePhasedX),
    "kak_decomposition": Tket(TketOptimizations.KAKDecomposition),
    "one": Tket(TketOptimizations.One),
    "peephole_optimise_2q": Tket(TketOptimizations.PeepholeOptimise2Q),
    "remove_barriers": Tket(TketOptimizations.RemoveBarriers),
    "remove_discarded": Tket(TketOptimizations.RemoveDiscarded),
    "remove_redundancies": Tket(TketOptimizations.RemoveRedundancies),
    "simplify_measured": Tket(TketOptimizations.SimplifyMeasured),
    "three_qubit_squash": Tket(TketOptimizations.ThreeQubitSquash),
    "two": Tket(TketOptimizations.Two),
}


class OQCDevice(QuantumDevice):
    """Device class for OQC devices."""

    def __init__(
        self, profile: qbraid.runtime.TargetProfile, client: qcaas_client.client.OQCClient
    ):
        super().__init__(profile=profile)
        self._client = client

    @property
    def client(self) -> qcaas_client.client.OQCClient:
        """Returns the client for the device."""
        return self._client

    def __str__(self):
        """String representation of the OQCDevice object."""
        return f"{self.__class__.__name__}('{self.profile.device_name}')"

    def queue_depth(self) -> int:
        """Returns the number of tasks in the queue for the device."""
        try:
            exec_estimates = self._client.get_qpu_execution_estimates(qpu_ids=self.id)
            return exec_estimates["qpu_wait_times"][0]["tasks_in_queue"]
        except Exception as err:  # pylint: disable=broad-exception-caught
            raise ResourceNotFoundError("Queue depth is not available for this device.") from err

    def status(self) -> DeviceStatus:
        """Returns the status of the device."""
        feature_set: dict = self.profile.get("feature_set", {})
        always_on: bool = feature_set.get("always_on", False)
        if always_on:
            return DeviceStatus.ONLINE

        devices = self._client.get_qpus()
        device: Optional[dict] = next((d for d in devices if d["id"] == self.id), None)
        if not device:
            raise ResourceNotFoundError(f"Device '{self.id}' not found.")

        status: str = device.get("status", "")

        if status and status.upper() == "INACTIVE":
            return DeviceStatus.OFFLINE

        try:
            start_time = self.get_next_window()
            now = datetime.datetime.now()

            if now > start_time:  # TODO: does this comparison correctly account for timezones?
                return DeviceStatus.ONLINE
        except ResourceNotFoundError as err:  # pylint: disable=broad-exception-caught
            logger.info(err)

        return DeviceStatus.UNAVAILABLE

    def get_next_window(self) -> datetime.datetime:
        """
        Returns the start time of the next active window for the device.

        Note: Currently only AWS windows are defined.
        """
        try:
            start_time = self._client.get_next_window(self.id)
        except Exception as next_window_err:  # pylint: disable=broad-exception-caught
            try:
                exec_estimates = self._client.get_qpu_execution_estimates(qpu_ids=self.id)
                start_time = exec_estimates["qpu_wait_times"][0]["windows"][0]["start_time"]
            except Exception as exec_est_error:  # pylint: disable=broad-exception-caught
                logger.error(exec_est_error)
                raise ResourceNotFoundError(
                    f"Falied to fetch next active window for device '{self.id}'. "
                    "Note: Currently only AWS windows are defined."
                ) from next_window_err

        return datetime.datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S")

    def transform(self, run_input: str) -> str:
        """Transforms the input program before submitting it to the device."""
        qasm_module = pyqasm.loads(run_input)
        qasm_module.remove_includes()
        qasm_no_includes = pyqasm.dumps(qasm_module)
        return qasm_no_includes

    @staticmethod
    def _build_compiler_config(**kwargs) -> CompilerConfig:
        """Builds a compiler configuration object from the provided kwargs."""
        configs = {
            "shots",
            "repeats",
            "repetition_period",
            "results_format",
            "metrics",
            "active_calibrations",
            "optimizations",
            "error_mitigation",
        }

        unsupported_keys = set(kwargs) - configs
        if unsupported_keys:
            raise ValueError(f"Unsupported keyword arguments: {', '.join(unsupported_keys)}")

        default_values = {
            "shots": None,
            "repeats": None,
            "repetition_period": None,
            "results_format": "binary",
            "metrics": "default",
            "active_calibrations": None,
            "optimizations": None,
            "error_mitigation": None,
        }

        config_values = {key: kwargs.get(key, default) for key, default in default_values.items()}

        try:
            return CompilerConfig(
                repeats=config_values["shots"] or config_values["repeats"],
                repetition_period=config_values["repetition_period"],
                results_format=RESULTS_FORMAT[config_values["results_format"]],
                metrics=METRICS_TYPE[config_values["metrics"]],
                active_calibrations=config_values["active_calibrations"],
                optimizations=(
                    OPTIMIZATIONS[config_values["optimizations"]]
                    if config_values["optimizations"]
                    else None
                ),
                error_mitigation=config_values["error_mitigation"],
            )
        except KeyError as err:
            raise ValueError(f"Invalid configuration option: {err.args[0]}") from err

    # pylint: disable-next=arguments-differ
    def submit(self, run_input, **kwargs) -> Union[OQCJob, list[OQCJob]]:
        """Submit one or more jobs to the device."""
        is_single_input = not isinstance(run_input, list)
        run_input = [run_input] if is_single_input else run_input
        config = self._build_compiler_config(**kwargs) if any(kwargs) else None
        tasks = [QPUTask(program=program, config=config, qpu_id=self.id) for program in run_input]
        qpu_tasks = self._client.schedule_tasks(tasks, qpu_id=self.id)
        jobs = [OQCJob(job_id=task.task_id, device=self, client=self._client) for task in qpu_tasks]
        return jobs[0] if is_single_input else jobs



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining Amazon Braket AHS conversion extras.

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from braket.ahs import AnalogHamiltonianSimulation

from qbraid.transpiler.annotations import requires_extras

if TYPE_CHECKING:
    import bloqade.analog.builder.assign


@requires_extras("bloqade")
def bloqade_to_braket_ahs(
    program: bloqade.analog.builder.assign.BatchAssign,
) -> list[AnalogHamiltonianSimulation]:
    """Converts a Bloqade program batch to a list of Amazon Braket AHS.

    Args:
        program (BatchAssign): Bloqade program batch

    Returns:
        list[AnalogHamiltonianSimulation]: A list of Amazon Braket AHS programs
    """
    # pylint: disable=import-outside-toplevel
    from bloqade.analog.compiler.passes.hardware import (
        analyze_channels,
        assign_circuit,
        canonicalize_circuit,
        generate_ahs_code,
        generate_braket_ir,
        validate_waveforms,
    )

    braket_device_route = program.braket
    braket_hardware_routine = braket_device_route.aquila()

    circuit, params = braket_hardware_routine.circuit, braket_hardware_routine.params

    capabilities = braket_hardware_routine.backend.get_capabilities(use_experimental=False)

    task_specs = []

    for _, batch_params in enumerate(params.batch_assignments()):
        assignments = {**batch_params, **params.static_params}
        final_circuit, _ = assign_circuit(circuit, assignments)

        level_couplings = analyze_channels(final_circuit)
        final_circuit = canonicalize_circuit(final_circuit, level_couplings)

        validate_waveforms(level_couplings, final_circuit)
        ahs_components = generate_ahs_code(capabilities, level_couplings, final_circuit)

        task_ir = generate_braket_ir(ahs_components, 0)
        ahs_program = AnalogHamiltonianSimulation.from_ir(task_ir.program)

        task_specs.append(ahs_program)

    return task_specs



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for QbraidDevice, QbraidJob, and QbraidGateModelResultBuilder
classes using the qbraid_qir_simulator

"""
from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from ._resources import MockDevice

if TYPE_CHECKING:
    import pyqir
    from pyqir import Module

pyqir = pytest.importorskip("pyqir")


@pytest.fixture
def pyqir_module() -> Module:
    """Returns a one-qubit PyQIR module with Hadamard gate and measurement."""
    bell = pyqir.SimpleModule("test_qir_program", num_qubits=1, num_results=1)
    qis = pyqir.BasicQisBuilder(bell.builder)

    qis.h(bell.qubits[0])
    qis.mz(bell.qubits[0], bell.results[0])

    return bell._module


def test_device_transform(pyqir_module: Module, mock_qbraid_device):
    """Test transform method on OpenQASM 2 string."""
    assert mock_qbraid_device.prepare(pyqir_module) == {"bitcode": pyqir_module.bitcode}


def test_transform_to_ir_from_spec(mock_basic_device: MockDevice, pyqir_module: Module):
    """Test transforming to run input to given IR from target profile program spec."""
    run_input_transformed = mock_basic_device.transform(pyqir_module)
    run_input_ir = mock_basic_device.prepare(run_input_transformed)
    assert isinstance(run_input_ir, dict)
    assert isinstance(run_input_ir.get("bitcode"), bytes)

    mock_basic_device._target_spec = None
    run_input_transformed = mock_basic_device.transform(pyqir_module)
    run_input_ir = mock_basic_device.prepare(run_input_transformed)
    assert isinstance(run_input_ir, pyqir.Module)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for IonQProvider class

"""
import importlib.util
import textwrap
import uuid
from itertools import combinations
from typing import Any, Optional
from unittest.mock import ANY, Mock, call, patch

import pytest
import qiskit
from packaging.version import parse
from qiskit import QuantumCircuit

from qbraid.passes.qasm import normalize_qasm_gate_params, rebase
from qbraid.programs import NATIVE_REGISTRY, ProgramSpec
from qbraid.programs.gate_model.ionq import IONQ_NATIVE_GATES_BASE, IONQ_QIS_GATES, GateSet
from qbraid.programs.typer import IonQDict
from qbraid.runtime import GateModelResultData, ResourceNotFoundError, Result, TargetProfile
from qbraid.runtime.enums import DeviceStatus, JobStatus
from qbraid.runtime.ionq import IonQDevice, IonQJob, IonQProvider, IonQSession
from qbraid.runtime.ionq.job import IonQJobError

qiskit_ge_v2 = parse(qiskit.__version__) >= parse("2.0.0")

FIXTURE_COUNT = sum(key in NATIVE_REGISTRY for key in ["qiskit", "braket", "cirq"])

DEVICE_DATA = [
    {
        "backend": "qpu.harmony",
        "status": "available",
        "qubits": 11,
        "average_queue_time": 1352464532,
        "last_updated": 1717253790,
        "has_access": False,
        "characterization_url": "/characterizations/ab9ff65b-693f-417f-a07b-927b3e48f6e6",
        "degraded": False,
    },
    {
        "backend": "qpu.aria-1",
        "status": "available",
        "qubits": 25,
        "average_queue_time": 136598879,
        "last_updated": 1717253790,
        "has_access": False,
        "characterization_url": "/characterizations/8c04e925-76fe-4ff6-83f3-9daf643abbf0",
        "degraded": False,
    },
    {
        "backend": "qpu.aria-2",
        "status": "available",
        "qubits": 25,
        "average_queue_time": 1449545843,
        "last_updated": 1717253790,
        "has_access": False,
        "characterization_url": "/characterizations/b652694a-912c-47a8-8a67-f6edfd3c6bb2",
        "degraded": False,
    },
    {
        "backend": "qpu.forte-1",
        "status": "available",
        "qubits": 36,
        "average_queue_time": 1533413000,
        "last_updated": 1717253790,
        "has_access": False,
        "characterization_url": "/characterizations/4e8623ac-d592-4763-a3b6-ad8de775b735",
        "degraded": False,
    },
    {
        "backend": "simulator",
        "status": "available",
        "qubits": 29,
        "average_queue_time": 0,
        "last_updated": 1717253790,
        "has_access": True,
        "noise_models": ["aria-1", "harmony", "ideal"],
        "degraded": False,
    },
    {
        "backend": "qpu.forte-enterprise-1",
        "status": "unavailable",
        "qubits": 36,
        "average_queue_time": 2333081000,
        "last_updated": 1738692095,
        "has_access": False,
        "degraded": False,
    },
    {
        "backend": "qpu.forte-enterprise-2",
        "status": "unavailable",
        "qubits": 36,
        "average_queue_time": 0,
        "last_updated": 1738692095,
        "has_access": False,
        "degraded": False,
    },
]

POST_JOB_RESPONSE = {
    "id": "c86a043a-6aea-47cf-b3a6-70ab1e538cab",
    "status": "ready",
    "request": 1717311751,
}

GET_JOB_RESPONSE = {
    "id": "c86a043a-6aea-47cf-b3a6-70ab1e538cab",
    "submitted_by": "e1254d93988e405e80d7842a",
    "status": "completed",
    "target": "simulator",
    "qubits": 1,
    "circuits": 1,
    "results_url": "/v0.3/jobs/c86a043a-6aea-47cf-b3a6-70ab1e538cab/results",
    "gate_counts": {"1q": 1, "2q": 0},
    "cost_usd": 0,
    "project_id": "0c2ac5b4-6078-40a0-9574-91f7aeab0b7e",
    "request": 1717296868,
    "start": 1717296869,
    "response": 1717296869,
    "execution_time": 39,
    "predicted_execution_time": 2,
    "noise": {"model": "ideal"},
    "error_mitigation": {"debias": False},
    "children": [],
}

GET_JOB_RESULT_RESPONSE = {"0": 0.5, "1": 0.5}

CHARACTERIZATION_DATA = {
    "connectivity": list(combinations(range(11), 2)),
    "qubits": 11,
    "fidelity": {"1q": {"mean": 0.9973}, "spam": {"mean": 0.993}, "2q": {"mean": 0.9002}},
    "timing": {
        "1q": 1e-05,
        "reset": 2e-05,
        "t1": 10000,
        "readout": 0.00013,
        "t2": 0.2,
        "2q": 0.0002,
    },
    "date": 1725102011,
    "id": "f518d0c9-34c6-4854-890e-a0e4f339bd64",
    "backend": "qpu.harmony",
}


def assert_qasm_equal(qasm1, qasm2):
    """Assert that two QASM strings are equal."""
    assert textwrap.dedent(qasm1).strip() == textwrap.dedent(qasm2).strip()


def mock_characterization(device_data: dict[str, Any]) -> Optional[dict[str, Any]]:
    """Mock get characterization method."""
    if device_data["backend"] == CHARACTERIZATION_DATA["backend"]:
        return CHARACTERIZATION_DATA.copy()
    return None


@pytest.fixture
def program_spec():
    """Return a ProgramSpec instance for IonQ."""
    return [ProgramSpec(str, alias="qasm2"), ProgramSpec(str, alias="qasm3")]


def test_ionq_provider_get_device(program_spec):
    """Test getting IonQ provider and device."""
    with (
        patch("qbraid.runtime.ionq.provider.Session") as mock_session,
        patch(
            "qbraid.runtime.ionq.provider.IonQProvider._get_characterization"
        ) as mock_get_characterization,
    ):

        mock_get_characterization.side_effect = mock_characterization
        mock_session.return_value.get.return_value.json.return_value = DEVICE_DATA

        provider = IonQProvider(api_key="fake_api_key")
        assert isinstance(provider, IonQProvider)
        assert isinstance(provider.session, IonQSession)
        assert provider.session.api_key == "fake_api_key"
        assert provider.session.base_url == "https://api.ionq.co/v0.3"
        assert provider.session.headers["Content-Type"] == "application/json"
        assert provider.session.headers["Authorization"] == "apiKey fake_api_key"

        test_devices = provider.get_devices()
        assert isinstance(test_devices, list)
        for test_device in test_devices:
            assert isinstance(test_device, IonQDevice)
            assert test_device.profile["device_id"] in [device["backend"] for device in DEVICE_DATA]
            assert test_device.profile["simulator"] is False or test_device.id == "simulator"
            assert test_device.profile["num_qubits"] in [device["qubits"] for device in DEVICE_DATA]
            assert test_device.profile["program_spec"] == program_spec

        test_device = provider.get_device("qpu.harmony")
        assert isinstance(test_device, IonQDevice)
        assert test_device.profile["device_id"] == "qpu.harmony"
        assert test_device.profile["simulator"] is False
        assert test_device.profile["num_qubits"] == 11
        assert test_device.profile["program_spec"] == program_spec
        assert test_device.profile["basis_gates"] == set(IONQ_QIS_GATES + IONQ_NATIVE_GATES_BASE)
        assert test_device.profile["characterization"] == CHARACTERIZATION_DATA


def test_get_device_characterization_from_data():
    """Test getting fetching device characterization from characterization url in device data."""
    with patch("qbraid.runtime.ionq.provider.IonQSession") as mock_session:
        mock_session.return_value.get.return_value.json.return_value = CHARACTERIZATION_DATA

        provider = IonQProvider(api_key="fake_api_key")
        harmony_data = DEVICE_DATA[0]
        characterization = provider._get_characterization(harmony_data)
        assert characterization == CHARACTERIZATION_DATA


def test_provider_session_device_not_found_error():
    """Test that a ValueError is raised if the device is not found."""
    with patch("qbraid.runtime.ionq.provider.Session") as mock_session:
        mock_session.return_value.get.return_value.json.return_value = DEVICE_DATA

        provider = IonQProvider(api_key="fake_api_key")

        with pytest.raises(ResourceNotFoundError, match="Device 'fake_device' not found."):
            provider.get_device("fake_device")


def test_ionq_provider_device_unavailable():
    """Test getting IonQ provider and different status devices."""

    class MockSession:
        """Mock session class."""

        def get_device(self, device_id: str):
            """Mock get_device method."""
            res = DEVICE_DATA[0]
            if device_id == "qpu.harmony":
                res["status"] = "retired"
            elif device_id == "qpu.aria-1":
                res["status"] = "offline"
            elif device_id == "qpu.aria-2":
                res["status"] = "available"
            elif device_id == "simulator":
                res["status"] = "available"
            elif device_id == "qpu.forte-1":
                res["status"] = "unavailable"
            elif device_id.startswith("qpu.forte-enterprise"):
                res["status"] = "unavailable"
            elif device_id == "fake_device":
                res["status"] = "fake_status"
            return res

    unavailable_profile = TargetProfile(device_id="qpu.forte-1", simulator=False)
    unavailable_device = IonQDevice(unavailable_profile, MockSession())
    assert unavailable_device.status() == DeviceStatus.UNAVAILABLE

    offline_profile = TargetProfile(device_id="qpu.aria-1", simulator=False)
    offline_device = IonQDevice(offline_profile, MockSession())
    assert offline_device.status() == DeviceStatus.OFFLINE

    available_profile = TargetProfile(device_id="qpu.aria-2", simulator=False)
    available_device = IonQDevice(available_profile, MockSession())
    assert available_device.status() == DeviceStatus.ONLINE

    retired_profile = TargetProfile(device_id="qpu.harmony", simulator=False)
    retired_device = IonQDevice(retired_profile, MockSession())
    assert retired_device.status() == DeviceStatus.RETIRED

    fake_profile = TargetProfile(device_id="fake_device", simulator=False)
    fake_device = IonQDevice(fake_profile, MockSession())
    with pytest.raises(ValueError):
        fake_device.status()


@pytest.fixture
def qis_input_decomp():
    """Return a QIS gateset input."""
    return {
        "format": "ionq.circuit.v0",
        "gateset": "qis",
        "qubits": 2,
        "circuit": [
            {"gate": "h", "target": 0},
            {"gate": "h", "target": 1},
            {"gate": "ry", "target": 1, "rotation": 0.39269908169872414},
            {"gate": "cnot", "control": 0, "target": 1},
            {"gate": "ry", "target": 1, "rotation": -0.39269908169872414},
            {"gate": "cnot", "control": 0, "target": 1},
        ],
    }


@pytest.fixture
def qis_input():
    """Return a QIS gateset input."""
    return {
        "format": "ionq.circuit.v0",
        "gateset": "qis",
        "qubits": 2,
        "circuit": [
            {"gate": "h", "target": 0},
            {"gate": "h", "target": 1},
            {"gate": "ry", "control": 0, "target": 1, "rotation": 0.7853981633974483},
        ],
    }


@pytest.fixture
def native_input():
    """Return a native gateset input."""
    return {
        "format": "ionq.circuit.v0",
        "gateset": "native",
        "qubits": 3,
        "circuit": [
            {"gate": "ms", "targets": [0, 1], "phases": [0, 0]},
            {"gate": "ms", "targets": [1, 2], "phases": [-0.5, 0.6], "angle": 0.1},
            {"gate": "gpi", "phase": 0, "target": 0},
            {"gate": "gpi2", "phase": 0.2, "target": 1},
        ],
    }


def test_ionq_device_transform_run_input(qis_input, qis_input_decomp):
    """Test transforming OpenQASM 2 string to supported gates + json format."""
    qasm_input = """
    OPENQASM 2.0;
    qreg q[2];
    h q[0];
    h q[1];
    cry(pi/4) q[0], q[1];
    """

    with (
        patch("qbraid.runtime.ionq.provider.Session") as mock_session,
        patch(
            "qbraid.runtime.ionq.provider.IonQProvider._get_characterization"
        ) as mock_get_characterization,
    ):

        mock_get_characterization.side_effect = mock_characterization
        mock_session.return_value.get.return_value.json.return_value = DEVICE_DATA

        provider = IonQProvider(api_key="fake_api_key")
        test_devices = provider.get_devices()
        device = test_devices[0]
        qasm_compat = device.transform(qasm_input)
        program_json = device.prepare(qasm_compat)
        assert program_json == qis_input

        qasm_rebased = rebase(qasm_input, gateset={"h", "ry", "cx"}, gate_mappings={"cx": "cnot"})
        qasm_decomp = normalize_qasm_gate_params(qasm_rebased)
        program_json_decomp = device.prepare(qasm_decomp)
        assert program_json_decomp == qis_input_decomp

        dummy_provider = IonQProvider(api_key="fake_api_key")
        assert provider == dummy_provider


def test_ionq_device_transform_retry():
    """Test transforming OpenQASM 3 string through try and except block."""
    qasm_input = """
    OPENQASM 3.0;
    include "stdgates.inc";
    gate rccx _gate_q_0, _gate_q_1, _gate_q_2 {
    u2(0, pi) _gate_q_2;
    u1(pi/4) _gate_q_2;
    cx _gate_q_1, _gate_q_2;
    u1(-pi/4) _gate_q_2;
    cx _gate_q_0, _gate_q_2;
    u1(pi/4) _gate_q_2;
    cx _gate_q_1, _gate_q_2;
    u1(-pi/4) _gate_q_2;
    u2(0, pi) _gate_q_2;
    }
    qubit[3] q;
    rccx q[0], q[1], q[2];
    """

    qasm_out = """
    OPENQASM 3.0;
    include "stdgates.inc";
    qubit[3] q;
    rz(3.141592653589793) q[2];
    rx(1.5707963267948966) q[2];
    rz(4.71238898038469) q[2];
    rx(1.5707963267948966) q[2];
    rz(3.141592653589793) q[2];
    h q[2];
    rx(0.7853981633974483) q[2];
    h q[2];
    cnot q[1], q[2];
    h q[2];
    rx(-0.7853981633974483) q[2];
    h q[2];
    cnot q[0], q[2];
    h q[2];
    rx(0.7853981633974483) q[2];
    h q[2];
    cnot q[1], q[2];
    h q[2];
    rx(-0.7853981633974483) q[2];
    h q[2];
    rz(3.141592653589793) q[2];
    rx(1.5707963267948966) q[2];
    rz(4.71238898038469) q[2];
    rx(1.5707963267948966) q[2];
    rz(3.141592653589793) q[2];
    """

    with (
        patch("qbraid.runtime.ionq.provider.Session") as mock_session,
        patch(
            "qbraid.runtime.ionq.provider.IonQProvider._get_characterization"
        ) as mock_get_characterization,
        patch("qbraid.runtime.ionq.device.logger") as mock_logger,
    ):
        mock_get_characterization.side_effect = mock_characterization
        mock_session.return_value.get.return_value.json.return_value = DEVICE_DATA

        provider = IonQProvider(api_key="fake_api_key")
        test_devices = provider.get_devices()
        device = test_devices[0]
        qasm_compat = device.transform(qasm_input)
        assert_qasm_equal(qasm_compat, qasm_out)

        mock_logger.debug.assert_has_calls(
            [
                call("Failed to transform OpenQASM program for IonQ: %s", ANY),
                call("Retrying using pyqasm.unroll()..."),
            ]
        )
        assert mock_logger.debug.call_count == 2


@pytest.mark.parametrize("circuit", range(FIXTURE_COUNT), indirect=True)
@patch("qbraid_core.sessions.Session.get")
@patch("qbraid_core.sessions.Session.post")
def test_ionq_device_run_submit_job(mock_post, mock_get, circuit, monkeypatch):
    """Test running a fake job."""
    monkeypatch.setattr("qbraid.runtime.ionq.device.importlib.util.find_spec", lambda _: None)

    mock_get_response = Mock()
    mock_get_response.json.side_effect = [
        DEVICE_DATA,  # provider.get_device("simulator")
        DEVICE_DATA,  # ionq_device.run(circuit, shots=2)
        GET_JOB_RESPONSE,  # job.status()
        GET_JOB_RESPONSE,  # job.metadata()
        GET_JOB_RESPONSE,  # job.metadata()
        GET_JOB_RESPONSE,  # job.result()
        GET_JOB_RESULT_RESPONSE,  # job.result()
    ]
    mock_get.return_value = mock_get_response

    # Setup mock for post
    mock_post_response = Mock()
    mock_post_response.json.return_value = POST_JOB_RESPONSE
    mock_post.return_value = mock_post_response

    provider = IonQProvider(api_key="fake_api_key")
    ionq_device = provider.get_device("simulator")
    job = ionq_device.run(circuit, shots=2)
    assert isinstance(job, IonQJob)

    job_status = job.status()
    assert isinstance(job_status, JobStatus)
    assert job_status == JobStatus.COMPLETED

    def _validate_metadata(metadata):
        assert isinstance(metadata, dict)
        assert metadata["job_id"] == "c86a043a-6aea-47cf-b3a6-70ab1e538cab"
        assert metadata["shots"] == 2
        assert metadata["status"] == JobStatus.COMPLETED

    job_metadata = job.metadata()
    _validate_metadata(job_metadata)

    # test caching
    job_metadata = job.metadata()
    _validate_metadata(job_metadata)

    res = job.result()
    assert isinstance(res, Result)
    assert isinstance(res.data, GateModelResultData)
    assert res.data.get_counts() == {"0": 1, "1": 1}
    assert res.data.measurements is None


@pytest.mark.parametrize("circuit", range(FIXTURE_COUNT), indirect=True)
@patch("qbraid_core.sessions.Session.get")
@patch("qbraid_core.sessions.Session.post")
@patch("importlib.util.find_spec", return_value=None)
def test_ionq_failed_job(mock_find_spec, mock_post, mock_get, circuit):
    """Test running a fake job."""
    GET_JOB_RESPONSE["status"] = "failed"
    mock_get_response = Mock()
    mock_get_response.json.side_effect = [
        DEVICE_DATA,  # provider.get_device("simulator")
        DEVICE_DATA,  # ionq_device.run(circuit, shots=2)
        {"status": "failed"},  # job.status()
        GET_JOB_RESPONSE,  # job.metadata()
        GET_JOB_RESPONSE,  # job.result()
        GET_JOB_RESULT_RESPONSE,  # job.result()
    ]
    mock_get.return_value = mock_get_response

    # Setup mock for post
    mock_post_response = Mock()
    mock_post_response.json.return_value = POST_JOB_RESPONSE
    mock_post.return_value = mock_post_response

    provider = IonQProvider(api_key="fake_api_key")
    ionq_device = provider.get_device("simulator")
    job = ionq_device.run(circuit, shots=2)
    assert isinstance(job, IonQJob)

    job_status = job.status()
    assert isinstance(job_status, JobStatus)
    assert job_status == JobStatus.FAILED

    with pytest.raises(IonQJobError):
        job.result()


def test_ionq_job_cancel():
    """Test cancelling a job."""

    class FakeSession:
        """Fake session class."""

        def cancel_job(self, job_id: str):  # pylint: disable=unused-argument
            """Fake cancel job."""
            return None

    job = IonQJob("fake_job_id", FakeSession())
    assert job.cancel() is None


def test_ionq_session_cancel():
    """Test cancelling a job."""
    with patch("qbraid_core.sessions.Session.put") as mock_put:
        mock_put.return_value.json.return_value = None

        session = IonQSession(api_key="fake_api_key")
        assert session.cancel_job("fake_job_id") is None


def test_ionq_average_queue_time():
    """Test getting the average queue time of an IonQDevice."""
    device = IonQDevice(
        TargetProfile(device_id="simulator", simulator=False),
        IonQSession("fake_api_key"),
    )
    with patch("qbraid_core.sessions.Session.get") as mock_get:
        mock_get.return_value.json.return_value = DEVICE_DATA

        assert device.avg_queue_time() == 0


def test_ionq_submit_fail():
    """Test submitting a job that fails."""
    circuit = """
    OPENQASM 2.0;
    qreg q[2];
    ry(pi/4) q[0];
    """
    device = IonQDevice(
        TargetProfile(device_id="simulator", simulator=False),
        IonQSession("fake_api_key"),
    )
    with patch("qbraid_core.sessions.Session.get") as mock_get:
        mock_get.return_value.json.return_value = DEVICE_DATA
        with patch("qbraid_core.sessions.Session.post") as mock_post:
            mock_post.return_value.json.return_value = {"error": "fake_error"}

            with pytest.raises(ValueError):
                device.run(circuit, shots=2)


def test_ionq_device_str_representation():
    """Test the string representation of an IonQDevice."""
    profile = TargetProfile(device_id="simulator", simulator=True)
    session = IonQSession("fake_api_key")
    device = IonQDevice(profile, session)
    assert str(device) == "IonQDevice('simulator')"


@pytest.mark.parametrize("result", [{"probabilities": {"0": 0.5, "1": 0.5}}, {"shots": 100}])
def test_get_counts_raises_value_error_for_missing_data(result):
    """Test that _get_counts raises a ValueError if shots or probabilities are missing."""
    with pytest.raises(ValueError) as excinfo:
        IonQJob._get_counts(result)
    assert "Missing shots or probabilities in result data." in str(excinfo.value)


@pytest.fixture
def mock_ionq_provider():
    """Return a mock IonQProvider instance."""
    return IonQProvider(api_key="mock_api_key")


@patch("builtins.hash", autospec=True)
def test_hash_method_creates_and_returns_hash(mock_hash, mock_ionq_provider):
    """Test that the __hash__ method creates and returns a hash."""
    mock_hash.return_value = 7777
    provider_instance = mock_ionq_provider
    result = provider_instance.__hash__()  # pylint:disable=unnecessary-dunder-call
    mock_hash.assert_called_once_with(("mock_api_key", "https://api.ionq.co/v0.3"))
    assert result == 7777
    assert provider_instance._hash == 7777


def test_hash_method_returns_existing_hash(mock_ionq_provider):
    """Test that the __hash__ method returns an existing hash."""
    provider_instance = mock_ionq_provider
    provider_instance._hash = 1234
    result = provider_instance.__hash__()  # pylint:disable=unnecessary-dunder-call
    assert result == 1234


def test_ionq_session_raises_for_no_api_key(monkeypatch):
    """Test that IonQSession raises an error if no API key is provided."""
    monkeypatch.setenv("IONQ_API_KEY", "")

    with pytest.raises(ValueError) as excinfo:
        IonQSession()
    assert "An IonQ API key is required to initialize the session." in str(excinfo.value)


def test_ionq_session_api_key_from_env_var(monkeypatch):
    """Test that IonQSession gets the API key from the IONQ_API_KEY environment variable."""
    api_key = uuid.uuid4().hex
    monkeypatch.setenv("IONQ_API_KEY", api_key)

    session = IonQSession()

    assert session.api_key == api_key


def test_squash_multicircuit_input_empty_batch():
    """Test with an empty batch_input to ensure ValueError is raised."""
    with pytest.raises(ValueError, match="run_input list cannot be empty."):
        IonQDevice._squash_multicircuit_input([])


def test_squash_multicircuit_input_inconsistent_format(native_input):
    """Test with inconsistent formats to ensure ValueError is raised."""
    batch_input = [
        native_input,
        {**native_input, "format": "ionq.circuit.v1"},
    ]
    with pytest.raises(
        ValueError, match="All run_inputs must have the same value for key 'format'."
    ):
        IonQDevice._squash_multicircuit_input(batch_input)


def test_squash_multicircuit_input_inconsistent_gateset(qis_input, native_input):
    """Test with inconsistent basis_gatess to ensure ValueError is raised."""
    batch_input = [
        qis_input,
        native_input,
    ]
    with pytest.raises(
        ValueError, match="All run_inputs must have the same value for key 'gateset'."
    ):
        IonQDevice._squash_multicircuit_input(batch_input)


def test_squash_multicircuit_input_missing_required_field(native_input):
    """Test with missing required fields to ensure ValueError is raised."""
    batch_input = [
        {**native_input, "qubits": None},
        native_input,
    ]
    with pytest.raises(ValueError, match="All run_inputs must be an instance of ~IonQDict."):
        IonQDevice._squash_multicircuit_input(batch_input)


def test_squash_multicircuit_input_valid_single_entry(native_input):
    """Test with a valid single entry in batch_input."""
    batch_input = [native_input]
    result = IonQDevice._squash_multicircuit_input(batch_input)
    assert result == {
        "format": native_input["format"],
        "gateset": native_input["gateset"],
        "qubits": native_input["qubits"],
        "circuits": [{"circuit": native_input["circuit"], "name": "Circuit 0"}],
    }


def test_squash_multicircuit_input_valid_multiple_entries(native_input):
    """Test with multiple valid entries in batch_input."""
    modified_native_input = {**native_input, "qubits": 5, "circuit": native_input["circuit"]}
    batch_input = [native_input, modified_native_input]

    result = IonQDevice._squash_multicircuit_input(batch_input)
    assert result == {
        "format": native_input["format"],
        "gateset": native_input["gateset"],
        "qubits": 5,
        "circuits": [
            {"circuit": native_input["circuit"], "name": "Circuit 0"},
            {"circuit": modified_native_input["circuit"], "name": "Circuit 1"},
        ],
    }


def random_ionq_id(user=False):
    """Return a random IonQ ID."""
    hex = uuid.uuid4().hex

    if user:
        return hex[:24]

    return f"{hex[:8]}-{hex[8:12]}-{hex[12:16]}-{hex[16:20]}-{hex[20:]}"


def generate_job_data_ids(num_children=0):
    """Generate random IonQ job data IDs."""
    return {
        "id": random_ionq_id(),
        "submitted_by": random_ionq_id(user=True),
        "project_id": random_ionq_id(),
        "children": [random_ionq_id() for _ in range(num_children)],
    }


@pytest.fixture
def multicircuit_job_data():
    """Return mock data for a multicircuit job."""
    ids = generate_job_data_ids(num_children=2)

    return {
        **ids,
        "status": "completed",
        "target": "simulator",
        "type": "circuit",
        "qubits": 2,
        "circuits": 2,
        "results_url": f"/v0.3/jobs/{ids['id']}/results",
        "gate_counts": {"1q": 2, "2q": 2},
        "cost_usd": 0,
        "request": 1731088388,
        "start": 1731088390,
        "response": 1731088390,
        "execution_time": 98,
        "predicted_execution_time": 8,
        "noise": {"model": "ideal"},
        "error_mitigation": {"debias": False},
        "probabilities": {
            ids["children"][0]: {"0": 0.65, "3": 0.35},
            ids["children"][1]: {"0": 0.5, "3": 0.5},
        },
        "shots": 1000,
    }


def test_ionq_get_counts_multicircuit_job(multicircuit_job_data):
    """Test getting counts for a multi-circuit job."""
    counts = IonQJob._get_counts(multicircuit_job_data)
    assert counts == [
        {"00": 650, "11": 350},
        {"00": 500, "11": 500},
    ]


@pytest.fixture
def qiskit_circuit():
    """Return a Qiskit circuit with U gate."""
    qc = QuantumCircuit(1)
    qc.u(0.1, 0.2, 0.3, 0)
    return qc


@pytest.mark.skipif(
    importlib.util.find_spec("qiskit_ionq") is None or qiskit_ge_v2,
    reason="qiskit-ionq not available.",
)
@pytest.mark.parametrize("gateset", ["native", "qis"])
def test_qiskit_ionq_conversion_type(qiskit_circuit, gateset):
    """Test that the output of the qiskit_ionq conversion is an IonQDict."""
    device = IonQDevice(
        TargetProfile(device_id="simulator", simulator=True),
        IonQSession("fake_api_key"),
    )
    with patch("qbraid_core.sessions.Session.get") as mock_get:
        mock_get.return_value.json.return_value = DEVICE_DATA
    output = device._apply_qiskit_ionq_conversion([qiskit_circuit], gateset=gateset)[0]
    assert isinstance(output, IonQDict)


@pytest.mark.skipif(
    importlib.util.find_spec("qiskit_ionq") is None or qiskit_ge_v2,
    reason="qiskit-ionq not available.",
)
@pytest.mark.parametrize(
    "gateset,expected",
    [
        (
            "native",
            {
                "format": "ionq.circuit.v0",
                "gateset": "native",
                "qubits": 1,
                "circuit": [
                    {"gate": "gpi2", "target": 0, "phase": 0.4522535170724314},
                    {"gate": "gpi", "target": 0, "phase": -3.469446951953614e-18},
                    {"gate": "gpi2", "target": 0, "phase": 0.5318309886183791},
                ],
            },
        ),
        (
            "qis",
            {
                "format": "ionq.circuit.v0",
                "gateset": "qis",
                "qubits": 1,
                "circuit": [
                    {"gate": "rz", "targets": [0], "rotation": 0.2999999999999994},
                    {"gate": "ry", "targets": [0], "rotation": 0.10000000000000005},
                    {"gate": "rz", "targets": [0], "rotation": 0.19999999999999973},
                ],
            },
        ),
    ],
)
def test_qiskit_ionq_conversion_output(qiskit_circuit, gateset, expected):
    """Test the output of the qiskit_ionq conversion is the expected IonQDict."""
    device = IonQDevice(
        TargetProfile(device_id="simulator", simulator=True),
        IonQSession("fake_api_key"),
    )
    with patch("qbraid_core.sessions.Session.get") as mock_get:
        mock_get.return_value.json.return_value = DEVICE_DATA
    output = device._apply_qiskit_ionq_conversion([qiskit_circuit], gateset=gateset)[0]
    assert output == expected


def test_ionq_device_run_warnings(monkeypatch):
    """Test that appropriate warnings are raised when using "
    qiskit-specific parameters with non-qiskit input."""
    monkeypatch.setattr("qbraid.runtime.ionq.device.importlib.util.find_spec", lambda _: None)

    device = IonQDevice(
        TargetProfile(device_id="simulator", simulator=True),
        IonQSession("fake_api_key"),
    )

    device.submit = Mock()

    dummy_circuit = "OPENQASM 2.0; qreg q[1]; h q[0];"

    with pytest.warns(UserWarning, match="GateSet argument is only applicable when qiskit-ionq"):
        device.run(dummy_circuit, shots=1000, gateset=GateSet.QIS)

    with pytest.warns(UserWarning, match="IonQ compiler synthesis option is only applicable when"):
        device.run(dummy_circuit, shots=1000, ionq_compiler_synthesis=True)

    with pytest.warns() as record:
        device.run(dummy_circuit, shots=1000, gateset=GateSet.QIS, ionq_compiler_synthesis=True)

    assert len(record) == 2
    assert "GateSet argument is only applicable" in str(record[0].message)
    assert "IonQ compiler synthesis option is only applicable" in str(record[1].message)

    device.submit.assert_called()


@pytest.mark.skipif(
    importlib.util.find_spec("qiskit_ionq") is None or qiskit_ge_v2,
    reason="qiskit-ionq not available.",
)
def test_ionq_device_run_batch_with_qiskit_ionq(qiskit_circuit):
    """Test calling IonQDevice.run with batch of qiskit circuits
    and qiskit-ionq installed."""
    device = IonQDevice(
        TargetProfile(device_id="simulator", simulator=True),
        IonQSession("fake_api_key"),
    )
    with patch("qbraid_core.sessions.Session.get") as mock_get:
        mock_get.return_value.json.return_value = DEVICE_DATA

    device.submit = Mock(return_value=[Mock(), Mock()])

    qiskit_batch = [qiskit_circuit, qiskit_circuit]

    device.run(qiskit_batch, shots=1000, gateset=GateSet.QIS)

    device.submit.assert_called_once()



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name,unused-argument,too-many-lines

"""
Unit tests for QbraidDevice and QbraidProvider classes.

"""
import importlib.util
import json
import logging
import time
from typing import Any
from unittest.mock import Mock, patch

import cirq
import numpy as np
import pytest
from qbraid_core.services.quantum.exceptions import QuantumServiceRequestError

from qbraid._caching import cache_disabled
from qbraid.programs import (
    ExperimentType,
    ProgramSpec,
    register_program_type,
    unregister_program_type,
)
from qbraid.programs.exceptions import ProgramTypeError
from qbraid.programs.typer import IonQDict, QuboCoefficientsDict
from qbraid.runtime import DeviceStatus, JobStatus, Result, TargetProfile, ValidationLevel
from qbraid.runtime.exceptions import QbraidRuntimeError, ResourceNotFoundError
from qbraid.runtime.native import QbraidDevice, QbraidJob, QbraidProvider
from qbraid.runtime.native.provider import _serialize_sequence, get_program_spec_lambdas
from qbraid.runtime.native.result import NECVectorAnnealerResultData, QbraidQirSimulatorResultData
from qbraid.runtime.noise import NoiseModel
from qbraid.runtime.options import RuntimeOptions
from qbraid.runtime.schemas.experiment import QuboSolveParams
from qbraid.runtime.schemas.job import RuntimeJobModel
from qbraid.transpiler import Conversion, ConversionGraph, ConversionScheme, ProgramConversionError

from ._resources import DEVICE_DATA_QIR, JOB_DATA_NEC, JOB_DATA_QIR, RESULTS_DATA_NEC, MockDevice

# Skip pulser tests if not installed
pulser_found = importlib.util.find_spec("pulser") is not None

if pulser_found:
    from .azure.test_azure_remote import (  # pylint: disable=unused-import # noqa: F401
        pulser_sequence,
    )


@pytest.fixture
def mock_nec_va_profile():
    """Mock profile for testing."""
    return TargetProfile(
        device_id="nec_vector_annealer",
        simulator=True,
        experiment_type=ExperimentType.ANNEALING,
        program_spec=QbraidProvider._get_program_spec("qubo", "nec_vector_annealer"),
    )


@pytest.fixture
def mock_scheme():
    """Mock conversion scheme for testing."""
    conv1 = Conversion("alice", "bob", lambda x: x + 1)
    conv2 = Conversion("bob", "charlie", lambda x: x - 1)
    graph = ConversionGraph(conversions=[conv1, conv2])
    scheme = ConversionScheme(conversion_graph=graph)
    return scheme


@pytest.fixture
def mock_qbraid_device(mock_profile, mock_scheme, mock_client):
    """Mock QbraidDevice for testing."""
    return QbraidDevice(profile=mock_profile, client=mock_client, scheme=mock_scheme)


@pytest.fixture
def mock_nec_va_device(mock_nec_va_profile, mock_client):
    """Mock NEC vector annealer device."""
    return QbraidDevice(profile=mock_nec_va_profile, client=mock_client)


def counts_to_measurements(counts: dict[str, Any]) -> np.ndarray:
    """Convert counts dictionary to measurements array."""
    measurements = []
    for state, count in counts.items():
        measurements.extend([list(map(int, state))] * count)
    return np.array(measurements, dtype=int)


def is_uniform_comput_basis(array: np.ndarray) -> bool:
    """
    Check if each measurement (row) in the array represents a uniform computational basis
    state, i.e., for each shot, that qubit measurements are either all |0⟩s or all |1⟩s.

    Args:
        array (np.ndarray): A 2D numpy array where each row represents a measurement shot,
                            and each column represents a qubit's state in that shot.

    Returns:
        bool: True if every measurement is in a uniform computational basis state
              (all |0⟩s or all |1⟩s). False otherwise.

    Raises:
        ValueError: If the given array is not 2D.
    """
    if array.ndim != 2:
        raise ValueError("The input array must be 2D.")

    for shot in array:
        # Check if all qubits in the shot are measured as |0⟩ or all as |1⟩
        if not (np.all(shot == 0) or np.all(shot == 1)):
            return False
    return True


def test_provider_equality(mock_provider, mock_client):
    """Test equality of provider instances."""
    assert mock_provider == QbraidProvider(client=mock_client)
    assert mock_provider != mock_client


def test_qbraid_device_str_representation(mock_qbraid_device):
    """Test string representation of QbraidDevice."""
    assert str(mock_qbraid_device) == "QbraidDevice('qbraid_qir_simulator')"


@pytest.mark.skipif(importlib.util.find_spec("pyqir") is None, reason="pyqir is not installed")
def test_qir_simulator_workflow(mock_provider, cirq_uniform):
    """Test qir simulator qbraid device job submission and result retrieval."""
    circuit = cirq_uniform(num_qubits=5)
    num_qubits = len(circuit.all_qubits())

    provider = mock_provider
    device = provider.get_device("qbraid_qir_simulator")

    shots = 10
    job = device.run(circuit, shots=shots)
    assert isinstance(job, QbraidJob)
    assert job.is_terminal_state()

    batch_job = device.run([circuit], shots=shots, noise_model=NoiseModel("ideal"))
    assert isinstance(batch_job, list)
    assert all(isinstance(job, QbraidJob) for job in batch_job)

    result = job.result()
    assert isinstance(result, Result)
    assert isinstance(result.data, QbraidQirSimulatorResultData)
    assert repr(result.data).startswith("QbraidQirSimulatorResultData")
    assert result.success
    assert result.job_id == JOB_DATA_QIR["qbraidJobId"]
    assert result.device_id == JOB_DATA_QIR["qbraidDeviceId"]

    counts = result.data.get_counts()
    probabilities = result.data.get_probabilities()
    assert len(counts) == len(probabilities) == 2
    assert sum(probabilities.values()) == 1.0
    assert is_uniform_comput_basis(result.data.measurements)

    assert result.details["shots"] == shots
    assert result.details["metadata"]["circuitNumQubits"] == num_qubits
    assert isinstance(result.details["timeStamps"]["executionDuration"], int)

    with pytest.warns(DeprecationWarning, match=r"Call to deprecated function measurement_counts*"):
        assert result.measurement_counts() == counts

    with pytest.warns(DeprecationWarning, match=r"Call to deprecated function measurements*"):
        assert is_uniform_comput_basis(result.measurements())


def test_nec_vector_annealer_workflow(mock_provider):
    """Test NEC Vector Annealer job submission and result retrieval."""
    provider = mock_provider
    device = provider.get_device("nec_vector_annealer")
    coefficients = {("x1", "x1"): 3.0, ("x1", "x2"): 2.0}
    quadratic = {json.dumps(key): value for key, value in coefficients.items()}
    payload = {"problem": json.dumps({"quadratic": quadratic, "offset": 0.0})}
    job = device.submit(run_input=payload)
    assert isinstance(job, QbraidJob)
    assert job.is_terminal_state()

    result = job.result()
    assert isinstance(result, Result)
    assert isinstance(result.data, NECVectorAnnealerResultData)
    assert result.success
    assert result.job_id == JOB_DATA_NEC["qbraidJobId"]
    assert result.device_id == JOB_DATA_NEC["qbraidDeviceId"]
    assert result.data._solutions == RESULTS_DATA_NEC["solutions"]


@pytest.mark.skipif(importlib.util.find_spec("pyqir") is None, reason="pyqir is not installed")
def test_run_forbidden_kwarg(mock_provider):
    """Test that invoking run method with forbidden kwarg raises value error."""
    circuit = Mock()
    provider = mock_provider
    device = provider.get_device("qbraid_qir_simulator")

    device.update_scheme(conversion_graph=ConversionGraph())

    with pytest.raises(ValueError):
        device.run(circuit, shots=10, num_qubits=5)


def test_device_update_scheme(mock_qbraid_device):
    """Test updating ConversionScheme."""
    graph = mock_qbraid_device.scheme.conversion_graph.copy()
    new_conversion = Conversion("charlie", "alice", lambda x: x)
    graph.add_conversion(new_conversion)

    assert graph != mock_qbraid_device.scheme.conversion_graph

    mock_qbraid_device.update_scheme(conversion_graph=graph)

    assert graph == mock_qbraid_device.scheme.conversion_graph


def test_device_noisey_run_raises_for_unsupported(mock_qbraid_device):
    """Test raising exception when noise model is not supported."""
    with pytest.raises(ValueError):
        mock_qbraid_device.run(Mock(), noise_model=NoiseModel("amplitude_damping"))


def test_device_extract_qasm(valid_qasm2, mock_qbraid_device):
    """Test that the extracting OpenQASM representation function
    returns qasm2 string for qasm2 program spec."""
    assert (
        mock_qbraid_device._extract_qasm_rep(valid_qasm2, ProgramSpec(str, "qasm2")) == valid_qasm2
    )


def test_device_extract_qasm_rep_none(mock_qbraid_device):
    """Test that the extracting OpenQASM representation function returns
    None if no OpenQASM conversion is supported from given program type."""
    try:
        alias = "unittest"
        register_program_type(Mock, alias, overwrite=True)
        assert mock_qbraid_device._extract_qasm_rep(None, ProgramSpec(Mock, alias)) is None
    finally:
        unregister_program_type(alias)


def test_device_run_raises_for_protected_kwargs(valid_qasm2, mock_qbraid_device):
    """Test raising exception when run method is invoked with protected kwargs."""
    kwargs = {"openQasm": valid_qasm2}
    with pytest.raises(ValueError):
        mock_qbraid_device.run(valid_qasm2, **kwargs)


def test_provider_initialize_client_raises_for_multiple_auth_params(mock_client):
    """Test raising exception when initializing client if provided
    both an api key and a client object."""
    with pytest.raises(ValueError):
        QbraidProvider(api_key="abc123", client=mock_client)


def test_provider_resource_not_found_error_for_bad_api_key():
    """Test raising ResourceNotFoundError when the client fails to authenticate."""
    provider = QbraidProvider(api_key="abc123")
    with pytest.raises(ResourceNotFoundError):
        _ = provider.client


@patch("qbraid.runtime.native.provider.QuantumClient")
def test_provider_client_from_valid_api_key(client, mock_client):
    """Test a valid API key."""
    client.return_value = mock_client
    provider = QbraidProvider(api_key="abc123")
    assert provider.client == mock_client


def test_provider_get_devices(mock_client):
    """Test getting devices from the client."""
    client = mock_client
    provider = QbraidProvider(client=client)
    devices = provider.get_devices(qbraid_id="qbraid_qir_simulator")
    assert len(devices) == 1
    assert devices[0].id == "qbraid_qir_simulator"


def test_provider_get_devices_raises_for_no_results(mock_client):
    """Test raising ResourceNotFoundError when no devices are found."""
    client = mock_client
    provider = QbraidProvider(client=client)
    with pytest.raises(ResourceNotFoundError, match="No devices found matching given criteria."):
        provider.get_devices(provider="IBM")


def test_provider_get_cached_devices(mock_client, device_data_qir, monkeypatch):
    """Test getting devices from the cache."""
    monkeypatch.setenv("DISABLE_CACHE", "0")

    data = device_data_qir.copy()

    client = mock_client
    provider = QbraidProvider(client=client)
    client.search_devices = Mock()
    client.search_devices.return_value = [data]

    _ = provider.get_devices(qbraid_id="qbraid_qir_simulator")

    # second call should be from cache
    _ = provider.get_devices(qbraid_id="qbraid_qir_simulator")
    client.search_devices.assert_called_once()

    provider.get_devices.cache_clear()


def test_provider_get_devices_post_cache_expiry(mock_client, device_data_qir, monkeypatch):
    """Test that the cache entry is invalidated when the cache is too old."""
    monkeypatch.setenv("DISABLE_CACHE", "0")
    monkeypatch.setenv("_QBRAID_TEST_CACHE_CALLS", "1")

    data = device_data_qir.copy()

    client = mock_client
    provider = QbraidProvider(client=client)
    client.search_devices = Mock()
    client.search_devices.return_value = [data]

    init_time = time.time()
    device_ttl = 120
    _ = provider.get_devices(qbraid_id="qbraid_qir_simulator")

    # second call should not come from cache
    with patch("time.time", return_value=init_time + device_ttl + 5):
        _ = provider.get_devices(qbraid_id="qbraid_qir_simulator")
    assert client.search_devices.call_count == 2


def test_provider_get_devices_bypass_cache(mock_client, device_data_qir, monkeypatch):
    """Test that the cache is bypassed when the bypass_cache flag is set."""
    monkeypatch.setenv("DISABLE_CACHE", "0")
    monkeypatch.setenv("_QBRAID_TEST_CACHE_CALLS", "1")

    data = device_data_qir.copy()

    client = mock_client
    provider = QbraidProvider(client=client)
    client.search_devices = Mock()
    client.search_devices.return_value = [data]

    with cache_disabled(provider):
        assert provider.__cache_disabled is True  # pylint: disable=no-member
        _ = provider.get_devices(qbraid_id="qbraid_qir_simulator")

    assert provider.__cache_disabled is False  # pylint: disable=no-member
    assert client.search_devices.call_count == 1
    provider.get_devices.cache_clear()


def test_provider_search_devices_raises_for_bad_client(mock_client):
    """Test raising ResourceNotFoundError when the client fails to authenticate."""
    provider = QbraidProvider(client=mock_client)
    with pytest.raises(ResourceNotFoundError):
        provider.get_devices(qbraid_id="qbraid_qir_simulator", status="Bad status")


def test_provider_program_spec_none():
    """Test getting program spec when it is None."""
    assert QbraidProvider._get_program_spec(None, "device_id") is None


def test_device_queue_depth_raises(mock_basic_device):
    """Test raising exception when queue depth is unavailable."""
    with pytest.raises(ResourceNotFoundError):
        mock_basic_device.queue_depth()


def test_device_queue_depth(mock_qbraid_device):
    """Test getting queue depth."""
    assert mock_qbraid_device.queue_depth() == 0


def test_device_status(mock_qbraid_device, mock_profile, mock_client):
    """Test getting device status."""
    assert mock_qbraid_device.status() == DeviceStatus.ONLINE

    original_data = DEVICE_DATA_QIR.copy()
    try:
        DEVICE_DATA_QIR["status"] = None
        with pytest.raises(QbraidRuntimeError):
            device = QbraidDevice(profile=mock_profile, client=mock_client)
            device.status()
    finally:
        DEVICE_DATA_QIR.clear()
        DEVICE_DATA_QIR.update(original_data)


def test_try_extracting_info_exception_handling(caplog, mock_qbraid_device):
    """Test try_extracting_info exception handling."""
    obj = mock_qbraid_device

    def problematic_func():
        raise ValueError("This is a test error")

    caplog.set_level(logging.INFO)
    result = obj.try_extracting_info(problematic_func, "Error encountered")
    assert result is None
    assert (
        "Error encountered: This is a test error. Field will be omitted in job metadata."
        in caplog.text
    )


def test_device_metadata(mock_basic_device):
    """Test getting device metadata."""
    metadata = mock_basic_device.metadata()
    assert metadata["device_id"] == "qbraid_qir_simulator"
    assert metadata["simulator"] is True
    assert metadata["num_qubits"] == 64


def test_failed_circuit_conversion(mock_basic_device, mock_scheme):
    """Test raising exception when circuit conversion fails."""

    class FailMockTypeA:
        """Mock type for testing."""

    class FailMockTypeB:
        """Mock type for testing."""

    fake_spec = ProgramSpec(FailMockTypeA, alias="alice")
    mock_basic_device._target_spec = ProgramSpec(FailMockTypeB, alias="charlie")
    mock_basic_device._scheme = mock_scheme
    mock_input = FailMockTypeA()
    with pytest.raises(ProgramConversionError):
        mock_basic_device.transpile(mock_input, fake_spec)

    mock_basic_device._target_spec = None
    assert mock_basic_device.transpile(mock_input, fake_spec) == mock_input

    unregister_program_type("alice")
    unregister_program_type("charlie")


def test_wrong_type_conversion(mock_basic_device):
    """Test raising exception when circuit conversion fails due to wrong type."""

    class MockTypeA:
        """Mock type for testing."""

        def __init__(self, x):
            self.x = x

        def __add__(self, other):
            return MockTypeA(self.x + other.x)

        def __eq__(self, other):
            return self.x == other.x

    class MockTypeB:
        """Mock type for testing."""

        def __init__(self, x):
            self.x = x

        def __add__(self, other):
            return MockTypeB(self.x + other.x)

        def __eq__(self, other):
            return self.x == other.x

    conv1 = Conversion("alice", "charlie", lambda x: x + MockTypeA(1))
    graph = ConversionGraph(conversions=[conv1])
    scheme = ConversionScheme(conversion_graph=graph)

    fake_spec = ProgramSpec(MockTypeA, alias="alice", overwrite=True)
    mock_basic_device._target_spec = ProgramSpec(MockTypeB, alias="charlie", overwrite=True)
    mock_basic_device._scheme = scheme
    mock_input = MockTypeA(1)
    with pytest.raises(ProgramConversionError):
        mock_basic_device.transpile(mock_input, fake_spec)

    unregister_program_type("alice")
    unregister_program_type("charlie")


class FakeSession:
    """Mock session for testing."""

    def save_config(self, **kwargs):  # pylint: disable=unused-argument
        """Mock save configuration."""
        return None


class FakeClient:
    """Mock client for testing."""

    def __init__(self, api_key=None):  # pylint: disable=unused-argument
        self.session = FakeSession()

    def get_device(self, qbraid_id):
        """Mock get device."""
        raise ValueError("Device not found")


def test_save_config():
    """Test saving configuration."""
    provider = QbraidProvider(client=FakeClient())
    try:
        provider.save_config(param1="value1", param2="value2")
    except Exception:  # pylint: disable=broad-except
        pytest.fail("save_config raised an exception unexpectedly")


def test_get_device_fail():
    """Test raising exception when device is not found."""
    provider = QbraidProvider(client=FakeClient())
    with pytest.raises(ResourceNotFoundError):
        provider.get_device("qbraid_qir_simulator")


def test_set_options(mock_qbraid_device: QbraidDevice):
    """Test updating the default runtime options."""
    default_options = {"transpile": True, "transform": True, "validate": 2, "prepare": True}
    assert dict(mock_qbraid_device._options) == default_options

    mock_qbraid_device.set_options(transform=False)
    updated_options = default_options.copy()
    updated_options["transform"] = False
    assert dict(mock_qbraid_device._options) == updated_options

    mock_qbraid_device.set_options(validate=False)
    assert mock_qbraid_device._options["validate"] == ValidationLevel.NONE


def test_set_options_raises_for_bad_key(mock_basic_device: MockDevice):
    """Test that the set options method raises AttributeError for key
    not already included in options."""
    with pytest.raises(AttributeError):
        mock_basic_device.set_options(bad_key=True)


def test_estimate_cost_success(mock_qbraid_device):
    """Test estimate_cost returns the correct cost."""
    mock_core_client = Mock()
    mock_core_client.estimate_cost.return_value = 8.75
    mock_qbraid_device._client = mock_core_client

    cost = mock_qbraid_device.estimate_cost(shots=100, execution_time=1.1)
    assert float(cost) == 8.75
    mock_core_client.estimate_cost.assert_called_once_with(mock_qbraid_device.id, 100, 1.1)


@pytest.mark.parametrize(
    "shots, execution_time",
    [
        (None, None),
        (-1, 1.0),
        (100, -1.0),
        (0, 0),
        (None, 0),
        (0, None),
    ],
)
def test_estimate_cost_raises_for_invalid_args(mock_qbraid_device, shots, execution_time):
    """Test that estimate_cost raises ValueError for invalid arguments."""
    with pytest.raises(ValueError):
        mock_qbraid_device.estimate_cost(shots=shots, execution_time=execution_time)


def test_estimate_cost_resource_not_found_error(mock_qbraid_device):
    """Test estimate_cost raises ResourceNotFoundError if the core client fails."""
    mock_core_client = Mock()
    mock_core_client.estimate_cost.side_effect = QuantumServiceRequestError("Request failed")
    mock_qbraid_device._client = mock_core_client

    with pytest.raises(QbraidRuntimeError):
        mock_qbraid_device.estimate_cost(shots=100, execution_time=10.0)


@pytest.mark.parametrize(
    "status, status_text",
    [("FAILED", "Custom status text"), (JobStatus.FAILED, "Different custom status text")],
)
def test_runtime_job_model_from_dict_custom_status(status, status_text):
    """Test creating a RuntimeJobModel with custom status."""
    job_data = JOB_DATA_QIR.copy()
    job_data["status"] = status
    job_data["statusText"] = status_text

    model = RuntimeJobModel.from_dict(job_data)

    assert model.status == JobStatus.FAILED
    assert model.status_text == status_text


@pytest.mark.skipif(importlib.util.find_spec("pyqubo") is None, reason="pyqubo is not installed")
def test_construct_aux_payload_annealing(mock_nec_va_device):
    """Test constructing auxiliary payload with an annealing program."""
    from pyqubo import Spin  # pylint: disable=import-outside-toplevel

    s1, s2, s3, s4 = Spin("s1"), Spin("s2"), Spin("s3"), Spin("s4")
    H = (4 * s1 + 2 * s2 + 7 * s3 + s4) ** 2
    model = H.compile()

    aux_payload = mock_nec_va_device._construct_aux_payload(model)
    assert len(aux_payload) == 1
    assert aux_payload["numVariables"] == 4


@pytest.fixture
def mock_qbraid_client():
    """Mock client for testing."""
    client = Mock()
    client._user_metadata = {"organization": "qbraid", "role": "guest"}
    client.session.api_key = "mock_api_key"
    return client


@patch("builtins.hash", autospec=True)
def test_hash_method_creates_and_returns_hash(mock_hash, mock_qbraid_client):
    """Test that the hash method creates and returns a hash."""
    mock_hash.return_value = 8888
    provider_instance = QbraidProvider(client=mock_qbraid_client)
    result = provider_instance.__hash__()  # pylint:disable=unnecessary-dunder-call
    expected_organization_role = "qbraid-guest"
    mock_hash.assert_called_once_with(
        ("QbraidProvider", "mock_api_key", expected_organization_role)
    )
    assert result == 8888
    assert provider_instance._hash == 8888


def test_hash_method_returns_existing_hash(mock_qbraid_client):
    """Test that the hash method returns an existing hash."""
    provider_instance = QbraidProvider(client=mock_qbraid_client)
    provider_instance._hash = 5678
    result = provider_instance.__hash__()  # pylint:disable=unnecessary-dunder-call
    assert result == 5678


def test_device_merge_options(mock_profile, mock_client):
    """ "Test merging default and custom options for a device."""
    options = RuntimeOptions(validate=0)
    device = QbraidDevice(profile=mock_profile, client=mock_client, options=options)

    default_options = device._default_options()
    assert device._options != default_options
    assert device._options != options
    assert default_options["validate"] == 2
    assert device._options["validate"] == 0

    device.set_options(validate=2)
    assert device._options == default_options


def test_device_validate_non_native_type_logs(mock_qbraid_device, caplog):
    """Test that validate method skips qubit count validation for non-native types."""
    mock_qbraid_device._target_spec = None

    with caplog.at_level(logging.INFO):
        mock_qbraid_device.validate([123])

    assert (
        "Skipping qubit count validation: program type 'int' not supported natively." in caplog.text
    )


def test_device_validate_level_none(mock_qbraid_device):
    """Test that validate returns immediately if validate level set to 0."""
    mock_qbraid_device.set_options(validate=0)

    with patch.object(mock_qbraid_device, "status") as mock_status:
        result = mock_qbraid_device.validate(["abc123"])
        assert result is None
        mock_status.assert_not_called()


def test_resolve_noise_model_raises_for_bad_input_type(mock_qbraid_device):
    """Test raising exception when given noise model is not a valid type."""
    with pytest.raises(ValueError) as excinfo:
        mock_qbraid_device._resolve_noise_model(10)
    assert "Invalid type for noise model" in str(excinfo)


def test_validate_run_input_payload_valid_dict():
    """Test when the payload is a valid dictionary."""
    payload = {"key": "value"}
    target_spec = ProgramSpec(str, "qasm2")

    assert QbraidDevice._validate_run_input_payload(payload, None) is None
    assert QbraidDevice._validate_run_input_payload(payload, target_spec) is None


def test_validate_run_input_payload_invalid_payload_none_target_spec():
    """Test when the payload is not a dictionary and target_spec is None."""
    payload = "invalid_payload"
    target_spec = None

    with pytest.raises(QbraidRuntimeError) as excinfo:
        QbraidDevice._validate_run_input_payload(payload, target_spec)

    assert "Run input transform failed due to missing target ProgramSpec" in str(excinfo.value)
    assert "Ensure all required dependency extras for this device are installed" in str(
        excinfo.value
    )


def test_validate_run_input_payload_invalid_payload_with_target_spec():
    """Test when the payload is not a dictionary and target_spec is provided."""
    payload = "invalid_payload"
    target_spec = ProgramSpec(str, "qasm2")

    with pytest.raises(QbraidRuntimeError) as excinfo:
        QbraidDevice._validate_run_input_payload(payload, target_spec)

    assert "Run input transform failed, likely due to corrupted target ProgramSpec" in str(
        excinfo.value
    )
    assert "Use QbraidProvider.get_device() to re-instantiate the device object" in str(
        excinfo.value
    )


def test_get_program_spec_not_registered_warning():
    """Test that a warning is emitted when the program type is not registered."""
    run_package = "not_registered"
    device_id = "fake_device"
    with pytest.warns(
        RuntimeWarning,
        match=(
            f"The default runtime configuration for device '{device_id}' includes "
            f"transpilation to program type '{run_package}', which is not registered."
        ),
    ):
        QbraidProvider._get_program_spec(run_package, device_id)


def test_get_program_spec_lambdas_validate_qasm_to_ionq():
    """Test that the validate lambda for qasm3 programs raises exception
    for CircuitConversion error though validate_qasm_to_ionq."""
    program_type_alias = "qasm3"
    device_id = "ionq_simulator"
    invalid_program = "invalid qasm3 code"

    lambdas = get_program_spec_lambdas(program_type_alias, device_id)
    validate = lambdas["validate"]

    with patch("qbraid.runtime.native.provider.transpile") as mock_convert:
        mock_convert.side_effect = ProgramConversionError("Invalid QASM3 code")

        with pytest.raises(
            ValueError,
            match=(
                f"OpenQASM programs submitted to the {device_id} "
                "must be compatible with IonQ JSON format."
            ),
        ):
            validate(invalid_program)

        mock_convert.assert_called_once_with(invalid_program, "ionq", max_path_depth=1)


def test_get_program_spec_lambdas_pulser():
    """Test that the validate lambda for pulser programs."""
    pytest.importorskip("pulser", reason="Pasqal pulser package is not installed.")
    program_type_alias = "pulser"
    device_id = "pasqal.sim.emu-tn"

    lambdas = get_program_spec_lambdas(program_type_alias, device_id)

    assert lambdas["validate"] is None
    assert lambdas["serialize"] is _serialize_sequence


@pytest.mark.skipif(not pulser_found, reason="pulser not installed")
def test_sequence_serializer(pulser_sequence):
    """Test the serialization of a Pasqal Pulser sequence."""
    pytest.importorskip("pulser", reason="Pasqal pulser package is not installed.")

    serialized = _serialize_sequence(pulser_sequence)

    assert "sequenceBuilder" in serialized
    assert serialized["sequenceBuilder"] == pulser_sequence.to_abstract_repr()


def test_provider_get_basis_gates_ionq():
    """Test getting basis gates for IonQ device."""
    device_data = {"provider": "IonQ", "objArg": "simulator"}
    basis_gates = QbraidProvider._get_basis_gates(device_data)
    native = {"gpi", "gpi2", "ms", "zz"}
    unique = set(basis_gates)
    assert len(unique) == len(basis_gates) == 35
    assert native.issubset(unique)


@pytest.mark.parametrize(
    "program_spec, target_program_type_expected",
    [
        (None, None),
        (ProgramSpec(str, "qasm2"), "qasm2"),
        ([ProgramSpec(str, "qasm2"), ProgramSpec(str, "qasm3")], ["qasm2", "qasm3"]),
    ],
)
def test_device_metadata_for_different_program_specs(program_spec, target_program_type_expected):
    """Test getting device metadata for different program specs types: none, single and multiple."""
    profile = TargetProfile(
        device_id="fake_device",
        simulator=True,
        experiment_type=ExperimentType.GATE_MODEL,
        num_qubits=42,
        program_spec=program_spec,
    )

    device = MockDevice(profile=profile)

    metadata = device.metadata()

    target_program_type = metadata["runtime_config"]["target_program_type"]
    assert (
        target_program_type is None
        if target_program_type_expected is None
        else set(target_program_type) == set(target_program_type_expected)
    )

    with pytest.raises(ProgramTypeError) as excinfo:
        device._get_target_spec(cirq.Circuit())
    assert "Could not find a target ProgramSpec matching the alias" in str(excinfo.value)


def test_device_transpile_program_conversion_error():
    """Test that transpile method handles program conversion error
    correctly for device with multiple program specs."""
    circuit = {
        "qubits": 3,
        "gateset": "qis",
        "circuit": [
            {"gate": "h", "target": 0},
            {"gate": "cnot", "control": 0, "target": 1},
            {"gate": "cnot", "control": 0, "target": 2},
        ],
    }

    profile = TargetProfile(
        device_id="fake_device",
        simulator=True,
        experiment_type=ExperimentType.GATE_MODEL,
        num_qubits=42,
        program_spec=[ProgramSpec(str, "qasm2"), ProgramSpec(str, "qasm3")],
    )

    device = MockDevice(profile=profile)
    with pytest.raises(ProgramConversionError) as excinfo:
        device.transpile(circuit, ProgramSpec(IonQDict, "ionq"))
    assert "Transpile step failed after multiple attempts." in str(excinfo.value)


@pytest.fixture
def mock_qubo_solve_params() -> dict[str, Any]:
    """Return a minimal set of QUBO solve parameters."""
    return {"offset": 0.5}


@pytest.fixture
def mock_qubo_params_with_defaults(mock_qubo_solve_params) -> dict[str, Any]:
    """Return a comprehensive set of all possible QUBO solve parameters."""
    return {
        "offset": mock_qubo_solve_params["offset"],
        "num_reads": 1,
        "num_results": 1,
        "num_sweeps": 500,
        "beta_range": (10.0, 100.0, 200),
        "vector_mode": "accuracy",
        "timeout": 1800,
    }


@pytest.mark.parametrize(
    "params",
    [
        {"offset": 0.5, "num_reads": 2},
        {"offset": 0.7, "num_results": 3},
    ],
)
def test_device_resolve_qubo_params(mock_nec_va_device: QbraidDevice, params: dict[str, Any]):
    """Test that resolve_qubo_params method returns the correct parameters from dict or model."""
    params_input = params.copy()
    resolved_params = mock_nec_va_device._resolve_qubo_params(params_input)
    assert resolved_params == params


def test_device_resolve_qubo_params_with_all_params(
    mock_nec_va_device: QbraidDevice,
    mock_qubo_solve_params,
    mock_qubo_params_with_defaults: dict[str, Any],
):
    """Test that the resolve_qubo_params method returns all parameters correctly."""
    params_model = QuboSolveParams(**mock_qubo_solve_params)
    resolved_params = mock_nec_va_device._resolve_qubo_params(params_model)
    filtered_params = {k: v for k, v in resolved_params.items() if v is not None}
    assert filtered_params == mock_qubo_params_with_defaults


@pytest.mark.parametrize(
    "device, params, expected_message",
    [
        (
            "mock_qbraid_device",
            {"offset": 0.5},
            "QUBO solve parameters are only supported for annealing devices.",
        ),
        ("mock_nec_va_device", "invalid_params", "Invalid type for QUBO solve parameters"),
    ],
)
def test_device_resolve_qubo_params_raises_exceptions(
    request: pytest.FixtureRequest,
    device: QbraidDevice,
    params: dict[str, Any],
    expected_message: str,
):
    """Test that the resolve_qubo_params method raises expected exceptions
    for invalid devices or parameter types."""
    device = request.getfixturevalue(device)
    with pytest.raises(ValueError) as excinfo:
        device._resolve_qubo_params(params)
    assert expected_message in str(excinfo.value)


@pytest.fixture
def qubo_coefficients() -> QuboCoefficientsDict:
    """Return a mock set of QUBO coefficients."""
    return {
        ("s1", "s1"): -160.0,
        ("s4", "s2"): 16.0,
        ("s3", "s1"): 224.0,
        ("s2", "s2"): -96.0,
        ("s4", "s1"): 32.0,
        ("s1", "s2"): 64.0,
        ("s3", "s2"): 112.0,
        ("s3", "s3"): -196.0,
        ("s4", "s4"): -52.0,
        ("s4", "s3"): 56.0,
    }


def test_device_call_resolve_params_from_run_method(
    mock_nec_va_device, qubo_coefficients, mock_qubo_solve_params
):
    """Test that the run method calls the resolve_qubo_params method
    with the provided parameters."""
    with (
        patch.object(mock_nec_va_device, "_resolve_qubo_params") as mock_resolve_qubo_params,
        patch.object(mock_nec_va_device, "submit") as mock_submit,
    ):
        mock_nec_va_device.run(qubo_coefficients, params=mock_qubo_solve_params)
        mock_resolve_qubo_params.assert_called_once_with(mock_qubo_solve_params)
        mock_submit.assert_called_once()


@pytest.fixture
def mock_qasm_profile():
    """Mock profile for testing."""
    return TargetProfile(
        device_id="mock_device",
        simulator=True,
        experiment_type=ExperimentType.GATE_MODEL,
        program_spec=[ProgramSpec(str, "qasm2"), ProgramSpec(str, "qasm3")],
    )


@pytest.fixture
def mock_qasm_device(mock_qasm_profile, mock_scheme, mock_client):
    """Mock QbraidDevice for testing."""
    return QbraidDevice(profile=mock_qasm_profile, client=mock_client, scheme=mock_scheme)


def test_set_target_program_type(mock_qbraid_device: QbraidDevice):
    """Test setting target spec."""
    target_spec = mock_qbraid_device._target_spec
    assert isinstance(target_spec, ProgramSpec)
    mock_qbraid_device.set_target_program_type(None)
    assert mock_qbraid_device._target_spec is None
    mock_qbraid_device.set_target_program_type(target_spec.alias)
    assert mock_qbraid_device._target_spec == target_spec


def test_set_target_program_type_multi_program_spec(mock_qasm_device: QbraidDevice):
    """Test setting target spec with multiple program specs."""
    target_spec = mock_qasm_device._target_spec
    assert isinstance(target_spec, list)
    assert all(isinstance(spec, ProgramSpec) for spec in target_spec)
    mock_qasm_device.set_target_program_type("qasm2")
    assert mock_qasm_device._target_spec.alias == "qasm2"
    mock_qasm_device.set_target_program_type("qasm3")
    assert mock_qasm_device._target_spec.alias == "qasm3"
    mock_qasm_device.set_target_program_type(["qasm2", "qasm3"])
    assert isinstance(mock_qasm_device._target_spec, list)
    assert all(isinstance(spec, ProgramSpec) for spec in mock_qasm_device._target_spec)
    assert len(mock_qasm_device._target_spec) == 2


def test_set_target_program_type_raises_for_multiple(mock_qasm_device: QbraidDevice):
    """Test setting target spec with multiple program specs."""
    with pytest.raises(ValueError):
        mock_qasm_device.set_target_program_type("qasm4")


def test_set_target_program_type_raises_for_single_invalid(mock_qbraid_device: QbraidDevice):
    """Test setting target spec raises ValueError for invalid alias."""
    with pytest.raises(ValueError):
        mock_qbraid_device.set_target_program_type("qasm4")


def test_set_target_program_type_raises_for_list_invalid(mock_qasm_device: QbraidDevice):
    """Test setting target spec raises if any alias in list is not in original spec."""
    with pytest.raises(ValueError):
        mock_qasm_device.set_target_program_type(["qasm2", "qasm4"])


def test_set_target_program_type_raises_for_list_if_spec_single(mock_qbraid_device: QbraidDevice):
    """Test setting target spec raises if list is given but spec expects single alias."""
    with pytest.raises(ValueError):
        mock_qbraid_device.set_target_program_type(["qasm2", "qasm4"])


def test_set_target_program_type_raises_for_duplicate(mock_qbraid_device: QbraidDevice):
    """Test setting target spec raises ValueError for duplicate alias input."""
    with pytest.raises(ValueError):
        mock_qbraid_device.set_target_program_type(["qasm3", "qasm3"])


@pytest.fixture
def mock_profile_program_spec_none():
    """Mock profile for testing with program_spec set to none"""
    return TargetProfile(
        device_id="mock_device",
        simulator=True,
        experiment_type=ExperimentType.GATE_MODEL,
        program_spec=None,
    )


@pytest.fixture
def mock_device_program_spec_none(mock_profile_program_spec_none, mock_client):
    """Mock QbraidDevice for testing with target profile program_spec set to None."""
    return QbraidDevice(profile=mock_profile_program_spec_none, client=mock_client)


def test_set_target_spec_raises_if_none(mock_device_program_spec_none: QbraidDevice):
    """Test setting target spec raises ValueError if program_spec is None."""
    with pytest.raises(ValueError):
        mock_device_program_spec_none.set_target_program_type("qasm2")



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining the Configuration class for quantum devices, providing necessary
parameters for integration with the qBraid runtime.

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Union

from pydantic import BaseModel, ConfigDict, field_validator, model_validator

from qbraid.programs import ExperimentType, ProgramSpec

from .noise import NoiseModelSet

if TYPE_CHECKING:
    from typing_extensions import Self


class TargetProfile(BaseModel):
    """
    Encapsulates read-only configuration settings for a quantum device, along with
    additional data necessary to formulate the runtime protocol(s) used to submit
    quantum jobs to the device within a specified environment for a given provider.

    """

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True, frozen=True)

    device_id: str
    simulator: bool
    experiment_type: Optional[ExperimentType] = None
    num_qubits: Optional[int] = None
    program_spec: Optional[Union[ProgramSpec, list[ProgramSpec]]] = None
    provider_name: Optional[str] = None
    basis_gates: Optional[Union[list[str], set[str], tuple[str, ...]]] = None
    noise_models: Optional[NoiseModelSet] = None

    @field_validator("basis_gates")
    @classmethod
    def validate_basis_gates(cls, value) -> Optional[set[str]]:
        """Validate the basis gates."""
        if value is None:
            return None

        return {gate.lower() for gate in value}

    @model_validator(mode="after")
    def validate_basis_gates_for_experiment_type(self) -> Self:
        """Validate the basis gates for the action type."""
        if self.basis_gates is not None and self.experiment_type != ExperimentType.GATE_MODEL:
            raise ValueError("basis_gates can only be specified for gate model experiments")
        return self

    def get(self, key: str, default: Any = None) -> Any:
        """Return the value of the key if it exists."""
        return getattr(self, key, default)

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key)

    def __len__(self) -> int:
        return sum(1 for v in self.__dict__.values() if v is not None)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining IonQProgram Class

"""
from __future__ import annotations

import json
from enum import Enum
from typing import Any

from qbraid.programs.exceptions import ProgramTypeError
from qbraid.programs.typer import IonQDict

from ._model import GateModelProgram

# https://docs.ionq.com/api-reference/v0.3/writing-quantum-programs#supported-gates
IONQ_QIS_GATES = [
    "x",
    "y",
    "z",
    "rx",
    "ry",
    "rz",
    "h",
    "not",
    "cnot",
    "s",
    "si",
    "t",
    "ti",
    "v",
    "vi",
    "swap",
    "cx",
    "cy",
    "cz",
    "crx",
    "cry",
    "crz",
    "ch",
    "ccnot",
    "cs",
    "csi",
    "ct",
    "cti",
    "cv",
    "cvi",
    "id",
    "zz",
]

IONQ_NATIVE_GATES_BASE = ["gpi", "gpi2"]

IONQ_NATIVE_GATES_FAMILY = {
    "aria": ["ms"] + IONQ_NATIVE_GATES_BASE,
    "forte": ["zz"] + IONQ_NATIVE_GATES_BASE,
}

# https://docs.ionq.com/api-reference/v0.3/native-gates-api#gates
IONQ_NATIVE_GATES = list(set().union(*IONQ_NATIVE_GATES_FAMILY.values()))


class GateSet(Enum):
    """Enumeration of IonQ gate sets types, native and qis (abstract)."""

    NATIVE = "native"
    QIS = "qis"


class InputFormat(Enum):
    """Enumeration of IonQ input format types."""

    CIRCUIT = "ionq.circuit.v0"
    QASM = "qasm"
    OPENQASM = "openqasm"
    QUIPPER = "quipper"


class IonQProgram(GateModelProgram):
    """Wrapper class for ``IonQDict`` objects."""

    def __init__(self, program: IonQDict):
        super().__init__(program)
        if not isinstance(program, IonQDict):
            raise ProgramTypeError(message=f"Expected 'IonQDict' object, got '{type(program)}'.")

    @property
    def qubits(self) -> list[int]:
        """Return the qubits acted upon by the operations in this circuit"""
        num_qubits: int = self.program["qubits"]
        return list(range(num_qubits))

    @property
    def num_clbits(self) -> int:
        """Return the number of classical bits in the circuit."""
        return 0

    @staticmethod
    def determine_gateset(circuit: list[dict[str, Any]]) -> GateSet:
        """Determines the gate set of an IonQ circuit gate list.

        Args:
            circuit (list[dict]): The IonQ circuit to analyze.

        Returns:
            GateSet: The gate set of the circuit.

        Raises:
            ValueError: If the circuit is empty, or mixes native and abstract (qis) gates.
        """
        if not circuit:
            raise ValueError("Circuit is empty. Must contain at least one gate.")

        native_gate_set = set(IONQ_NATIVE_GATES)

        def is_native_gate(gate_info: dict) -> bool:
            """Helper function to determine if a gate is native."""
            gate = gate_info.get("gate")
            if gate == "zz":
                return gate_info.get("angle") is not None
            return gate in native_gate_set

        first_gate_native = is_native_gate(circuit[0])

        for instr in circuit:
            if is_native_gate(instr) != first_gate_native:
                raise ValueError(
                    f"Invalid gate '{instr.get('gate')}'. "
                    "Cannot mix native and QIS gates in the same circuit."
                )

        return GateSet.NATIVE if first_gate_native else GateSet.QIS

    def validate_for_gateset(self) -> None:
        """Validate that the circuit only contains gates from the derived gate set.

        Raises:
            ValueError: If the circuit contains an invalid gate.
        """
        gate_set_map = {
            GateSet.NATIVE: set(IONQ_NATIVE_GATES),
            GateSet.QIS: set(IONQ_QIS_GATES),
        }

        circuit: list[dict[str, Any]] = self.program["circuit"]

        gate_set_name = IonQProgram.determine_gateset(circuit)

        gate_set = gate_set_map[gate_set_name]

        for instr in circuit:
            gate = instr.get("gate")
            if gate not in gate_set:
                raise ValueError(
                    f"Invalid gate '{gate}'. Must be in the '{gate_set_name.value}' gate set."
                )

    def serialize(self) -> dict[str, str]:
        """Return the program in a format suitable for submission to the qBraid API."""
        return {"ionqCircuit": json.dumps(self.program)}



# Copyright (C) 2024 qBraid
# Copyright (C) The Cirq Developers
#
# This file is part of the qBraid-SDK.
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. This specific file, adapted from Cirq, is dual-licensed under both the
# Apache License, Version 2.0, and the GPL v3. You may not use this file except in
# compliance with the applicable license. You may obtain a copy of the Apache License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file includes code adapted from Cirq (https://github.com/quantumlib/Cirq)
# with modifications by qBraid. The original copyright notice is included above.
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# qbraid: skip-header
# pylint: skip-file
# flake8: noqa
# fmt: off

"""
Module defining qBraid Cirq QASM parser.

"""
import functools
import operator
from typing import TYPE_CHECKING, Any, Callable, Iterable, Optional, Union, cast

import numpy as np
from cirq import CX, Circuit, NamedQubit, ops
from cirq.circuits.qasm_output import QasmUGate
from cirq.contrib.qasm_import._lexer import QasmLexer
from cirq.contrib.qasm_import.exception import QasmException
from qbraid_core._import import LazyLoader

from .cirq_custom import U2Gate, U3Gate, rzz

yacc = LazyLoader('yacc', globals(), 'ply.yacc')

# Redefined lexer tokens (4/7/21) to surpress warning:
# Token ['IF', 'NE'] defined, but not used
QasmLexer.tokens = [
    "FORMAT_SPEC",
    "NUMBER",
    "NATURAL_NUMBER",
    "QELIBINC",
    "ID",
    "PI",
    "QREG",
    "CREG",
    "MEASURE",
    "ARROW",
]

if TYPE_CHECKING:
    import cirq


class Qasm:
    """Qasm stores the final result of the Qasm parsing."""

    def __init__(
        self, supported_format: bool, qelib1_include: bool, qregs: dict, cregs: dict, c: Circuit
    ):
        # defines whether the Quantum Experience standard header
        # is present or not
        self.qelib1Include = qelib1_include
        # defines if it has a supported format or not
        self.supportedFormat = supported_format
        # circuit
        self.qregs = qregs
        self.cregs = cregs
        self.circuit = c


class QasmGateStatement:
    """Specifies how to convert a call to an OpenQASM gate
    to a list of cirq.GateOperation's.
    Has the responsibility to validate the arguments
    and parameters of the call and to generate a list of corresponding
    cirq.GateOperation's in the 'on' method.
    """

    def __init__(
        self,
        qasm_gate: str,
        cirq_gate: Union[ops.Gate, Callable[[list[float]], ops.Gate]],
        num_params: int,
        num_args: int,
    ):
        """Initializes a Qasm gate statement.
        Args:
            qasm_gate: The symbol of the QASM gate.
            cirq_gate: The gate class on the cirq side.
            num_params: The number of params taken by this gate.
            num_args: The number of qubits (used in validation) this gate takes.
        """
        self.qasm_gate = qasm_gate
        self.cirq_gate = cirq_gate
        self.num_params = num_params

        # at least one quantum argument is mandatory for gates to act on
        assert num_args >= 1
        self.num_args = num_args

    def _validate_args(self, args: list[list[ops.Qid]], lineno: int):
        if len(args) != self.num_args:
            raise QasmException(
                "{} only takes {} arg(s) (qubits and/or registers), "
                "got: {}, at line {}".format(self.qasm_gate, self.num_args, len(args), lineno)
            )

    def _validate_params(self, params: list[float], lineno: int):
        if len(params) != self.num_params:
            raise QasmException(
                "{} takes {} parameter(s), got: {}, at line {}".format(
                    self.qasm_gate, self.num_params, len(params), lineno
                )
            )

    def on(
        self, params: list[float], args: list[list[ops.Qid]], lineno: int
    ) -> Iterable[ops.Operation]:
        self._validate_args(args, lineno)
        self._validate_params(params, lineno)

        reg_sizes = np.unique([len(reg) for reg in args])
        if len(reg_sizes) > 2 or (len(reg_sizes) > 1 and reg_sizes[0] != 1):
            raise QasmException(
                f"Non matching quantum registers of length {reg_sizes} at line {lineno}"
            )

        # the actual gate we'll apply the arguments to might be a parameterized
        # or non-parameterized gate
        final_gate: ops.Gate = (
            self.cirq_gate if isinstance(self.cirq_gate, ops.Gate) else self.cirq_gate(params)
        )
        # OpenQASM gates can be applied on single qubits and qubit registers.
        # We represent single qubits as registers of size 1.
        # Based on the OpenQASM spec (https://arxiv.org/abs/1707.03429),
        # single qubit arguments can be mixed with qubit registers.
        # Given quantum registers of length reg_size and single qubits are both
        # used as arguments, we generate reg_size GateOperations via iterating
        # through each qubit of the registers 0 to n-1 and use the same one
        # qubit from the "single-qubit registers" for each operation.
        op_qubits = functools.reduce(
            cast(Callable[[list['cirq.Qid'], list['cirq.Qid']], list['cirq.Qid']], np.broadcast),
            args,
        )
        for qubits in op_qubits:
            if isinstance(qubits, ops.Qid):
                yield final_gate.on(qubits)
            elif len(np.unique(qubits)) < len(qubits):
                raise QasmException(f"Overlapping qubits in arguments at line {lineno}")
            else:
                yield final_gate.on(*qubits)


class QasmParser:
    """Cirq Parser for QASM strings.

    Example:
        qasm = "OPENQASM 2.0; qreg q1[2]; CX q1[0], q1[1];"
        parsedQasm = QasmParser().parse(qasm)
    """

    def __init__(self):
        self.parser = yacc.yacc(module=self, debug=False, write_tables=False)
        self.circuit = Circuit()
        self.qregs: dict[str, int] = {}
        self.cregs: dict[str, int] = {}
        self.qelibinc = False
        self.lexer = QasmLexer()
        self.supported_format = False
        self.parsedQasm: Optional[Qasm] = None
        self.qubits: dict[str, ops.Qid] = {}
        self.functions = {
            'sin': np.sin,
            'cos': np.cos,
            'tan': np.tan,
            'exp': np.exp,
            'ln': np.log,
            'sqrt': np.sqrt,
            'acos': np.arccos,
            'atan': np.arctan,
            'asin': np.arcsin,
        }

        self.binary_operators = {
            '+': operator.add,
            '-': operator.sub,
            '*': operator.mul,
            '/': operator.truediv,
            '^': operator.pow,
        }

    basic_gates: dict[str, QasmGateStatement] = {
        'CX': QasmGateStatement(qasm_gate='CX', cirq_gate=CX, num_params=0, num_args=2),
        'U': QasmGateStatement(
            qasm_gate='U',
            num_params=3,
            num_args=1,
            # QasmUGate expects half turns
            cirq_gate=(lambda params: QasmUGate(*[p / np.pi for p in params])),
        ),
    }

    qelib_gates = {
        'rx': QasmGateStatement(
            qasm_gate='rx', cirq_gate=(lambda params: ops.rx(params[0])), num_params=1, num_args=1
        ),
        'crx': QasmGateStatement(
            qasm_gate='crx', cirq_gate=(lambda params: ops.ControlledGate(ops.rx(params[0]))), num_params=1, num_args=2
        ),
        'sx': QasmGateStatement(
            qasm_gate='sx', num_params=0, num_args=1, cirq_gate=ops.XPowGate(exponent=0.5)
        ),
        'sxdg': QasmGateStatement(
            qasm_gate='sxdg', num_params=0, num_args=1, cirq_gate=ops.XPowGate(exponent=-0.5)
        ),
        'ry': QasmGateStatement(
            qasm_gate='ry', cirq_gate=(lambda params: ops.ry(params[0])), num_params=1, num_args=1
        ),
        'rz': QasmGateStatement(
            qasm_gate='rz', cirq_gate=(lambda params: ops.rz(params[0])), num_params=1, num_args=1
        ),
        'id': QasmGateStatement(
            qasm_gate='id', cirq_gate=ops.IdentityGate(1), num_params=0, num_args=1
        ),
        'u1': QasmGateStatement(
            qasm_gate='u1',
            cirq_gate=(lambda params: ops.ZPowGate(exponent=params[0] / np.pi)),
            num_params=1,
            num_args=1,
        ),
        'u2': QasmGateStatement(
            qasm_gate='u2',
            cirq_gate=(lambda params: U2Gate(*params)),
            num_params=2,
            num_args=1,
        ),
        'u3': QasmGateStatement(
            qasm_gate='u3',
            cirq_gate=(lambda params: U3Gate(*params)),
            num_params=3,
            num_args=1,
        ),
        'u': QasmGateStatement(
            qasm_gate='u',
            cirq_gate=(lambda params: U3Gate(*params)),
            num_params=3,
            num_args=1,
        ),
        'r': QasmGateStatement(
            qasm_gate='r',
            num_params=2,
            num_args=1,
            cirq_gate=(
                lambda params: QasmUGate(
                    params[0] / np.pi, (params[1] / np.pi) - 0.5, (-params[1] / np.pi) + 0.5
                )
            ),
        ),
        'x': QasmGateStatement(qasm_gate='x', num_params=0, num_args=1, cirq_gate=ops.X),
        'y': QasmGateStatement(qasm_gate='y', num_params=0, num_args=1, cirq_gate=ops.Y),
        'z': QasmGateStatement(qasm_gate='z', num_params=0, num_args=1, cirq_gate=ops.Z),
        'h': QasmGateStatement(qasm_gate='h', num_params=0, num_args=1, cirq_gate=ops.H),
        's': QasmGateStatement(qasm_gate='s', num_params=0, num_args=1, cirq_gate=ops.S),
        'cs': QasmGateStatement(qasm_gate='cs', num_params=0, num_args=2, cirq_gate=ops.ControlledGate(ops.S)),
        't': QasmGateStatement(qasm_gate='t', num_params=0, num_args=1, cirq_gate=ops.T),
        'cx': QasmGateStatement(qasm_gate='cx', cirq_gate=CX, num_params=0, num_args=2),
        'cy': QasmGateStatement(
            qasm_gate='cy', cirq_gate=ops.ControlledGate(ops.Y), num_params=0, num_args=2
        ),
        'cz': QasmGateStatement(qasm_gate='cz', cirq_gate=ops.CZ, num_params=0, num_args=2),
        'ccz': QasmGateStatement(qasm_gate='ccz', cirq_gate=ops.CCZ, num_params=0, num_args=3),
        'ch': QasmGateStatement(
            qasm_gate='ch', cirq_gate=ops.ControlledGate(ops.H), num_params=0, num_args=2
        ),
        'swap': QasmGateStatement(qasm_gate='swap', cirq_gate=ops.SWAP, num_params=0, num_args=2),
        'cswap': QasmGateStatement(
            qasm_gate='cswap', num_params=0, num_args=3, cirq_gate=ops.CSWAP
        ),
        'ccx': QasmGateStatement(qasm_gate='ccx', num_params=0, num_args=3, cirq_gate=ops.CCX),
        'c3x': QasmGateStatement(qasm_gate='c3x', num_params=0, num_args=4, cirq_gate=ops.ControlledGate(ops.CCX)),
        'c4x': QasmGateStatement(qasm_gate='c4x', num_params=0, num_args=5, cirq_gate=ops.ControlledGate(ops.ControlledGate(ops.CCX))),
        'sdg': QasmGateStatement(qasm_gate='sdg', num_params=0, num_args=1, cirq_gate=ops.ZPowGate(exponent=-0.5)),
        'csdg': QasmGateStatement(qasm_gate='csdg', num_params=0, num_args=2, cirq_gate=ops.ControlledGate(ops.ZPowGate(exponent=-0.5))),
        'tdg': QasmGateStatement(qasm_gate='tdg', num_params=0, num_args=1, cirq_gate=ops.ZPowGate(exponent=-0.25)),
        'crz': QasmGateStatement(
            qasm_gate='crz',
            cirq_gate=(lambda params: ops.ControlledGate(ops.rz(params[0]))),
            num_params=1,
            num_args=2,
        ),
        'cry': QasmGateStatement(
            qasm_gate='cry',
            cirq_gate=(lambda params: ops.ControlledGate(ops.ry(params[0]))),
            num_params=1,
            num_args=2
        ),
        'csx': QasmGateStatement(
            qasm_gate='csx', num_params=0, num_args=2, cirq_gate=ops.ControlledGate(ops.XPowGate(exponent=0.5))
        ),
        'c3sx': QasmGateStatement(
            qasm_gate='c3sx', num_params=0, num_args=4, cirq_gate=ops.ControlledGate(ops.ControlledGate(ops.ControlledGate(ops.XPowGate(exponent=0.5))))
        ),
        'c3sqrtx': QasmGateStatement(
            qasm_gate='c3sqrtx', num_params=0, num_args=4, cirq_gate=ops.ControlledGate(ops.ControlledGate(ops.ControlledGate(ops.XPowGate(exponent=0.5))))
        ),
        'cu1': QasmGateStatement(
            qasm_gate='cu1',
            cirq_gate=(lambda params: ops.ControlledGate(ops.ZPowGate(exponent=params[0] / np.pi))),
            num_params=1,
            num_args=2,
        ),
        'cu3': QasmGateStatement(
            qasm_gate='cu3',
            cirq_gate=(lambda params: ops.ControlledGate(U3Gate(*params))),
            num_params=3,
            num_args=2,
        ),
        'cu': QasmGateStatement(
            qasm_gate='cu',
            cirq_gate=(lambda params: ops.ControlledGate(U3Gate(*params))),
            num_params=3,
            num_args=2,
        ),
        'p': QasmGateStatement(
            qasm_gate='p', cirq_gate=(lambda params: ops.ZPowGate(exponent=params[0] / np.pi)), num_params=1, num_args=1,
        ),
        'cp': QasmGateStatement(
            qasm_gate='cp',
            cirq_gate=(lambda params: ops.CZPowGate(exponent=params[0] / np.pi)),
            num_params=1,
            num_args=2,
        ),
        'iswap': QasmGateStatement(
            qasm_gate='iswap', cirq_gate=ops.ISWAP, num_params=0, num_args=2
        ),
        'rzz': QasmGateStatement(
            qasm_gate='rzz',
            cirq_gate=(lambda params: rzz(params[0] / np.pi)),
            num_params=1,
            num_args=2,
        ),
    }

    all_gates = {**basic_gates, **qelib_gates}

    tokens = QasmLexer.tokens
    start = 'start'

    precedence = (('left', '+', '-'), ('left', '*', '/'), ('right', '^'))

    def p_start(self, p):
        """start : qasm"""
        p[0] = p[1]

    def p_qasm_format_only(self, p):
        """qasm : format"""
        self.supported_format = True
        p[0] = Qasm(self.supported_format, self.qelibinc, self.qregs, self.cregs, self.circuit)

    def p_qasm_no_format_specified_error(self, p):
        """qasm : QELIBINC
        | circuit"""
        if self.supported_format is False:
            raise QasmException("Missing 'OPENQASM 2.0;' statement")

    def p_qasm_include(self, p):
        """qasm : qasm QELIBINC"""
        self.qelibinc = True
        p[0] = Qasm(self.supported_format, self.qelibinc, self.qregs, self.cregs, self.circuit)

    def p_qasm_circuit(self, p):
        """qasm : qasm circuit"""
        p[0] = Qasm(self.supported_format, self.qelibinc, self.qregs, self.cregs, p[2])

    def p_format(self, p):
        """format : FORMAT_SPEC"""
        if p[1] != "2.0":
            raise QasmException(
                "Unsupported OpenQASM version: {}, "
                "only 2.0 is supported currently by Cirq".format(p[1])
            )

    # circuit : new_reg circuit
    #         | gate_op circuit
    #         | measurement circuit
    #         | if circuit
    #         | empty

    def p_circuit_reg(self, p):
        """circuit : new_reg circuit"""
        p[0] = self.circuit

    def p_circuit_gate_or_measurement(self, p):
        """circuit :  circuit gate_op
        |  circuit measurement"""
        self.circuit.append(p[2])
        p[0] = self.circuit

    def p_circuit_empty(self, p):
        """circuit : empty"""
        p[0] = self.circuit

    # qreg and creg

    def p_new_reg(self, p):
        """new_reg : QREG ID '[' NATURAL_NUMBER ']' ';'
        | CREG ID '[' NATURAL_NUMBER ']' ';'"""
        name, length = p[2], p[4]
        if name in self.qregs.keys() or name in self.cregs.keys():
            raise QasmException(f"{name} is already defined at line {p.lineno(2)}")
        if length == 0:
            raise QasmException(f"Illegal, zero-length register '{name}' at line {p.lineno(4)}")
        if p[1] == "qreg":
            self.qregs[name] = length
        else:
            self.cregs[name] = length
        p[0] = (name, length)

    # gate operations
    # gate_op : ID qargs
    #         | ID ( params ) qargs

    def p_gate_op_no_params(self, p):
        """gate_op :  ID qargs"""
        self._resolve_gate_operation(p[2], gate=p[1], p=p, params=[])

    def p_gate_op_with_params(self, p):
        """gate_op :  ID '(' params ')' qargs"""
        self._resolve_gate_operation(args=p[5], gate=p[1], p=p, params=p[3])

    def _resolve_gate_operation(
        self, args: list[list[ops.Qid]], gate: str, p: Any, params: list[float]
    ):
        gate_set = self.basic_gates if not self.qelibinc else self.all_gates
        if gate not in gate_set.keys():
            msg = 'Unknown gate "{}" at line {}{}'.format(
                gate,
                p.lineno(1),
                ", did you forget to include qelib1.inc?" if not self.qelibinc else "",
            )
            raise QasmException(msg)
        p[0] = gate_set[gate].on(args=args, params=params, lineno=p.lineno(1))

    # params : parameter ',' params
    #        | parameter

    def p_params_multiple(self, p):
        """params : expr ',' params"""
        p[3].insert(0, p[1])
        p[0] = p[3]

    def p_params_single(self, p):
        """params : expr"""
        p[0] = [p[1]]

    # expr : term
    #            | func '(' expression ')' """
    #            | binary_op
    #            | unary_op

    def p_expr_term(self, p):
        """expr : term"""
        p[0] = p[1]

    def p_expr_parens(self, p):
        """expr : '(' expr ')'"""
        p[0] = p[2]

    def p_expr_function_call(self, p):
        """expr : ID '(' expr ')'"""
        func = p[1]
        if func not in self.functions.keys():
            raise QasmException(f"Function not recognized: '{func}' at line {p.lineno(1)}")
        p[0] = self.functions[func](p[3])

    def p_expr_unary(self, p):
        """expr : '-' expr
        | '+' expr"""
        if p[1] == '-':
            p[0] = -p[2]
        else:
            p[0] = p[2]

    def p_expr_binary(self, p):
        """expr : expr '*' expr
        | expr '/' expr
        | expr '+' expr
        | expr '-' expr
        | expr '^' expr
        """
        p[0] = self.binary_operators[p[2]](p[1], p[3])

    def p_term(self, p):
        """term : NUMBER
        | NATURAL_NUMBER
        | PI"""
        p[0] = p[1]

    # qargs : qarg ',' qargs
    #      | qarg ';'

    def p_args_multiple(self, p):
        """qargs : qarg ',' qargs"""
        p[3].insert(0, p[1])
        p[0] = p[3]

    def p_args_single(self, p):
        """qargs : qarg ';'"""
        p[0] = [p[1]]

    # qarg : ID
    #     | ID '[' NATURAL_NUMBER ']'

    def p_quantum_arg_register(self, p):
        """qarg : ID"""
        reg = p[1]
        if reg not in self.qregs.keys():
            raise QasmException(f'Undefined quantum register "{reg}" at line {p.lineno(1)}')
        qubits = []
        for idx in range(self.qregs[reg]):
            arg_name = self.make_name(idx, reg)
            if arg_name not in self.qubits.keys():
                self.qubits[arg_name] = NamedQubit(arg_name)
            qubits.append(self.qubits[arg_name])
        p[0] = qubits

    # carg : ID
    #     | ID '[' NATURAL_NUMBER ']'

    def p_classical_arg_register(self, p):
        """carg : ID"""
        reg = p[1]
        if reg not in self.cregs.keys():
            raise QasmException(f'Undefined classical register "{reg}" at line {p.lineno(1)}')

        p[0] = [self.make_name(idx, reg) for idx in range(self.cregs[reg])]

    def make_name(self, idx, reg):
        return str(reg) + "_" + str(idx)

    def p_quantum_arg_bit(self, p):
        """qarg : ID '[' NATURAL_NUMBER ']'"""
        reg = p[1]
        idx = p[3]
        arg_name = self.make_name(idx, reg)
        if reg not in self.qregs.keys():
            raise QasmException(f'Undefined quantum register "{reg}" at line {p.lineno(1)}')
        size = self.qregs[reg]
        if idx >= size:
            raise QasmException(
                'Out of bounds qubit index {} '
                'on register {} of size {} '
                'at line {}'.format(idx, reg, size, p.lineno(1))
            )
        if arg_name not in self.qubits.keys():
            self.qubits[arg_name] = NamedQubit(arg_name)
        p[0] = [self.qubits[arg_name]]

    def p_classical_arg_bit(self, p):
        """carg : ID '[' NATURAL_NUMBER ']'"""
        reg = p[1]
        idx = p[3]
        arg_name = self.make_name(idx, reg)
        if reg not in self.cregs.keys():
            raise QasmException(f'Undefined classical register "{reg}" at line {p.lineno(1)}')

        size = self.cregs[reg]
        if idx >= size:
            raise QasmException(
                'Out of bounds bit index {} '
                'on classical register {} of size {} '
                'at line {}'.format(idx, reg, size, p.lineno(1))
            )
        p[0] = [arg_name]

    # measurement operations
    # measurement : MEASURE qarg ARROW carg

    def p_measurement(self, p):
        """measurement : MEASURE qarg ARROW carg ';'"""
        qreg = p[2]
        creg = p[4]

        if len(qreg) != len(creg):
            raise QasmException(
                'mismatched register sizes {} -> {} for measurement '
                'at line {}'.format(len(qreg), len(creg), p.lineno(1))
            )

        p[0] = [
            ops.MeasurementGate(num_qubits=1, key=creg[i]).on(qreg[i]) for i in range(len(qreg))
        ]

    # if operations
    # if : IF '(' carg EQ NATURAL_NUMBER ')' ID qargs

    # def p_if(self, p):
    #     """if : IF '(' carg EQ NATURAL_NUMBER ')' gate_op"""
    #     # We have to split the register into bits (since that's what measurement does above),
    #     # and create one condition per bit, checking against that part of the binary value.
    #     conditions = []
    #     for i, key in enumerate(p[3]):
    #         v = (p[5] >> i) & 1
    #         conditions.append(sympy.Eq(sympy.Symbol(key), v))
    #     p[0] = [
    #         ops.ClassicallyControlledOperation(conditions=conditions, sub_operation=tuple(p[7])[0])
    #     ]

    def p_error(self, p):
        if p is None:
            raise QasmException('Unexpected end of file')

        raise QasmException(
            f"""Syntax error: '{p.value}'
{self.debug_context(p)}
at line {p.lineno}, column {self.find_column(p)}"""
        )

    def find_column(self, p):
        line_start = self.qasm.rfind('\n', 0, p.lexpos) + 1
        return (p.lexpos - line_start) + 1

    def p_empty(self, p):
        """empty :"""

    def parse(self, qasm: str) -> Qasm:
        if self.parsedQasm is None:
            self.qasm = qasm
            self.lexer.input(self.qasm)
            self.parsedQasm = self.parser.parse(lexer=self.lexer)
        return self.parsedQasm

    def debug_context(self, p):
        debug_start = max(self.qasm.rfind('\n', 0, p.lexpos) + 1, p.lexpos - 5)
        debug_end = min(self.qasm.find('\n', p.lexpos, p.lexpos + 5), p.lexpos + 5)

        return (
            "..."
            + self.qasm[debug_start:debug_end]
            + "\n"
            + (" " * (3 + p.lexpos - debug_start))
            + "^"
        )
